"""Note: fixtures are stored in the tests/fixtures directory for better organization"""

from __future__ import annotations

import json
import os
import sys
from datetime import datetime, timedelta, timezone
from hashlib import md5
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, cast
from unittest import mock

import pytest
from click.testing import CliRunner
from filelock import FileLock
from git import Commit, Repo

from semantic_release.version.version import Version

from tests.const import PROJ_DIR
from tests.fixtures import *
from tests.util import copy_dir_tree, remove_dir_tree

if TYPE_CHECKING:
    from tempfile import _TemporaryFileWrapper
    from typing import Any, Callable, Generator, Optional, Protocol, Sequence, TypedDict

    from click.testing import Result
    from filelock import AcquireReturnProxy
    from git import Actor

    from tests.fixtures.git_repo import RepoActions

    class RunCliFn(Protocol):
        """
        Run the CLI with the provided arguments and a clean environment.

        :param argv: The arguments to pass to the CLI.
        :type argv: list[str] | None

        :param env: The environment variables to set for the CLI.
        :type env: dict[str, str] | None

        :param invoke_kwargs: Additional arguments to pass to the invoke method.
        :type invoke_kwargs: dict[str, Any] | None

        :return: The result of the CLI invocation.
        :rtype: Result
        """

        def __call__(
            self,
            argv: list[str] | None = None,
            env: dict[str, str] | None = None,
            invoke_kwargs: dict[str, Any] | None = None,
        ) -> Result: ...

    class MakeCommitObjFn(Protocol):
        def __call__(self, message: str) -> Commit: ...

    class NetrcFileFn(Protocol):
        def __call__(self, machine: str) -> _TemporaryFileWrapper[str]: ...

    class TeardownCachedDirFn(Protocol):
        def __call__(self, directory: Path) -> Path: ...

    class FormatDateStrFn(Protocol):
        def __call__(self, date: datetime) -> str: ...

    class GetStableDateNowFn(Protocol):
        def __call__(self) -> datetime: ...

    class GetMd5ForFileFn(Protocol):
        def __call__(self, file_path: Path | str) -> str: ...

    class GetMd5ForSetOfFilesFn(Protocol):
        """
        Generates a hash for a set of files based on their contents

        This function will automatically filter out any 0-byte files or `__init__.py` files

        :param: files: A list of file paths to generate a hash for (MUST BE absolute paths)
        """

        def __call__(self, files: Sequence[Path | str]) -> str: ...

    class GetAuthorizationToBuildRepoCacheFn(Protocol):
        def __call__(self, repo_name: str) -> AcquireReturnProxy | None: ...

    class BuildRepoOrCopyCacheFn(Protocol):
        def __call__(
            self,
            repo_name: str,
            build_spec_hash: str,
            build_repo_func: Callable[[Path], Sequence[RepoActions]],
            dest_dir: Path | None = None,
        ) -> Path: ...

    class RepoData(TypedDict):
        build_date: str
        build_spec_hash: str
        build_definition: Sequence[RepoActions]

    class GetCachedRepoDataFn(Protocol):
        def __call__(self, proj_dirname: str) -> RepoData | None: ...

    class SetCachedRepoDataFn(Protocol):
        def __call__(self, proj_dirname: str, data: RepoData) -> None: ...


def pytest_addoption(parser: pytest.Parser, pluginmanager: pytest.PytestPluginManager):
    parser.addoption(
        "--comprehensive",
        help="Run full test suite including slow tests",
        default=False,
        action="store_true",
    )


def pytest_configure(config: pytest.Config):
    """
    If no test selection modifications are provided, default to running only unit tests.

    See `pytest_collection_modifyitems` for more information on test selection modifications.
    """
    user_desired_comprehensive_evaluation = config.getoption("--comprehensive")
    user_provided_filter = str(config.getoption("-k"))
    user_provided_markers = str(config.getoption("-m"))

    root_test_dir = Path(__file__).parent.relative_to(config.rootpath)
    user_provided_test_path = bool(config.args != [str(root_test_dir)])

    # If no options are provided, default to running only unit tests
    if not any(
        (
            user_desired_comprehensive_evaluation,
            user_provided_test_path,
            user_provided_filter,
            user_provided_markers,
        )
    ):
        config.option.markexpr = pytest.mark.unit.name


@pytest.hookimpl(trylast=True)
def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]):
    """
    Test selection modifier based on markers and command line options.

    Examples
    --------
        pytest
            only unit tests that are not marked comprehensive are executed

        pytest --comprehensive
            all tests are executed

        pytest -m unit
            only unit tests that are not marked comprehensive are executed (same as no options)

        pytest -m e2e
            only end-to-end tests that are not marked comprehensive are executed

        pytest -m e2e --comprehensive
            all end-to-end tests are executed

        pytest -m "not unit"
            only tests that are not marked unit or comprehensive are executed

        pytest -m "not unit" --comprehensive
            all tests that are not marked unit are executed

        pytest -k "test_name"
            only tests that match the substring "test_name" (but not marked comprehensive) are executed

        pytest -k "test_name" --comprehensive
            all tests that match the substring "test_name" are executed

    """
    disable_comprehensive_tests = not config.getoption("--comprehensive")
    comprehensive_test_skip_marker = pytest.mark.skip(
        reason="comprehensive tests are disabled by default"
    )
    user_provided_filter = str(config.getoption("-k"))

    if any((disable_comprehensive_tests,)):
        for item in items:
            if user_provided_filter and user_provided_filter in item.name:
                continue
            if disable_comprehensive_tests and "comprehensive" in item.keywords:
                item.add_marker(comprehensive_test_skip_marker)


@pytest.fixture
def cli_runner() -> CliRunner:
    return CliRunner(mix_stderr=False)


@pytest.fixture(scope="session")
def run_cli(clean_os_environment: dict[str, str]) -> RunCliFn:
    def _run_cli(
        argv: list[str] | None = None,
        env: dict[str, str] | None = None,
        invoke_kwargs: dict[str, Any] | None = None,
    ) -> Result:
        from semantic_release.cli.commands.main import main
        from semantic_release.globals import logger

        # Prevent logs from being propagated to the root logger (pytest)
        logger.propagate = False

        cli_runner = CliRunner(mix_stderr=False)
        env_vars = {**clean_os_environment, **(env or {})}
        args = ["-vv", *(argv or [])]

        with mock.patch.dict(os.environ, env_vars, clear=True):
            # run the CLI with the provided arguments
            result = cli_runner.invoke(main, args=args, **(invoke_kwargs or {}))
            # Force the output to be printed to stdout which will be captured by pytest
            sys.stdout.write(result.stdout)
            sys.stderr.write(result.stderr)
            return result

    return _run_cli


@pytest.fixture(scope="session")
def default_netrc_username() -> str:
    return "username"


@pytest.fixture(scope="session")
def default_netrc_password() -> str:
    return "password"


@pytest.fixture(scope="session")
def netrc_file(
    default_netrc_username: str,
    default_netrc_password: str,
) -> Generator[NetrcFileFn, None, None]:
    temporary_files: list[str] = []

    def _netrc_file(machine: str) -> _TemporaryFileWrapper[str]:
        ctx_mgr = NamedTemporaryFile("w", delete=False)
        with ctx_mgr as netrc_fd:
            temporary_files.append(ctx_mgr.name)

            netrc_fd.write(f"machine {machine}{os.linesep}")
            netrc_fd.write(f"login {default_netrc_username}{os.linesep}")
            netrc_fd.write(f"password {default_netrc_password}{os.linesep}")
            netrc_fd.flush()
            return ctx_mgr

    try:
        yield _netrc_file
    finally:
        for temp_file in temporary_files:
            os.unlink(temp_file)


@pytest.fixture(scope="session")
def stable_today_date() -> datetime:
    curr_time = datetime.now(timezone.utc).astimezone()
    est_test_completion = curr_time + timedelta(hours=1)  # exaggeration
    starting_day_of_year = curr_time.timetuple().tm_yday
    ending_day_of_year = est_test_completion.timetuple().tm_yday

    if starting_day_of_year < ending_day_of_year:
        return est_test_completion

    return curr_time


@pytest.fixture(scope="session")
def stable_now_date(stable_today_date: datetime) -> GetStableDateNowFn:
    def _stable_now_date() -> datetime:
        curr_time = datetime.now(timezone.utc).astimezone()
        return stable_today_date.replace(
            minute=curr_time.minute,
            second=curr_time.second,
            microsecond=curr_time.microsecond,
        )

    return _stable_now_date


@pytest.fixture(scope="session")
def format_date_str() -> FormatDateStrFn:
    """Formats a date as how it would appear in the changelog (Must match local timezone)"""

    def _format_date_str(date: datetime) -> str:
        return date.strftime("%Y-%m-%d")

    return _format_date_str


@pytest.fixture(scope="session")
def today_date_str(
    stable_today_date: datetime, format_date_str: FormatDateStrFn
) -> str:
    """Today's Date formatted as how it would appear in the changelog (matches local timezone)"""
    return format_date_str(stable_today_date)


@pytest.fixture(scope="session")
def cached_files_dir(request: pytest.FixtureRequest) -> Path:
    return request.config.cache.mkdir("psr-cached-repos")


@pytest.fixture(scope="session")
def get_authorization_to_build_repo_cache(
    tmp_path_factory: pytest.TempPathFactory, worker_id: str
) -> GetAuthorizationToBuildRepoCacheFn:
    def _get_authorization_to_build_repo_cache(
        repo_name: str,
    ) -> AcquireReturnProxy | None:
        if worker_id == "master":
            # not executing with multiple workers via xdist, so just continue
            return None

        # get the temp directory shared by all workers
        root_tmp_dir = tmp_path_factory.getbasetemp().parent

        return FileLock(root_tmp_dir / f"{repo_name}.lock").acquire(
            timeout=30, blocking=True
        )

    return _get_authorization_to_build_repo_cache


@pytest.fixture(scope="session")
def get_cached_repo_data(request: pytest.FixtureRequest) -> GetCachedRepoDataFn:
    def _get_cached_repo_data(proj_dirname: str) -> RepoData | None:
        cache_key = f"psr/repos/{proj_dirname}"
        return cast("Optional[RepoData]", request.config.cache.get(cache_key, None))

    return _get_cached_repo_data


@pytest.fixture(scope="session")
def set_cached_repo_data(request: pytest.FixtureRequest) -> SetCachedRepoDataFn:
    def magic_serializer(obj: Any) -> Any:
        if isinstance(obj, Path):
            return obj.__fspath__()

        if isinstance(obj, Version):
            return obj.__dict__

        return obj

    def _set_cached_repo_data(proj_dirname: str, data: RepoData) -> None:
        cache_key = f"psr/repos/{proj_dirname}"
        request.config.cache.set(
            cache_key,
            json.loads(json.dumps(data, default=magic_serializer)),
        )

    return _set_cached_repo_data


@pytest.fixture(scope="session")
def build_repo_or_copy_cache(
    cached_files_dir: Path,
    today_date_str: str,
    stable_now_date: GetStableDateNowFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    set_cached_repo_data: SetCachedRepoDataFn,
    get_authorization_to_build_repo_cache: GetAuthorizationToBuildRepoCacheFn,
) -> BuildRepoOrCopyCacheFn:
    log_file = cached_files_dir.joinpath("repo-build.log")
    log_file_lock = FileLock(log_file.with_suffix(f"{log_file.suffix}.lock"), timeout=2)

    def _build_repo_w_cache_checking(
        repo_name: str,
        build_spec_hash: str,
        build_repo_func: Callable[[Path], Sequence[RepoActions]],
        dest_dir: Path | None = None,
    ) -> Path:
        # Blocking mechanism to synchronize xdist workers
        # Runs before the cache is checked because the cache will be set once the build is complete
        filelock = get_authorization_to_build_repo_cache(repo_name)

        cached_repo_data = get_cached_repo_data(repo_name)
        cached_repo_path = cached_files_dir.joinpath(repo_name)

        # Determine if the build spec has changed since the last cached build
        unmodified_build_spec = bool(
            cached_repo_data and cached_repo_data["build_spec_hash"] == build_spec_hash
        )

        if not unmodified_build_spec or not cached_repo_path.exists():
            # Cache miss, so build the repo (make sure its clean first)
            remove_dir_tree(cached_repo_path, force=True)
            cached_repo_path.mkdir(parents=True, exist_ok=True)

            build_msg = f"Building cached project files for {repo_name}"
            with log_file_lock, log_file.open(mode="a") as afd:
                afd.write(f"{stable_now_date().isoformat()}: {build_msg}...\n")

            try:
                # Try to build repository but catch any errors so that it doesn't cascade through all tests
                # do to an unreleased lock
                build_definition = build_repo_func(cached_repo_path)
            except Exception:
                remove_dir_tree(cached_repo_path, force=True)

                if filelock:
                    filelock.lock.release()

                with log_file_lock, log_file.open(mode="a") as afd:
                    afd.write(
                        f"{stable_now_date().isoformat()}: {build_msg}...FAILED\n"
                    )

                raise

            # Marks the date when the cached repo was created
            set_cached_repo_data(
                repo_name,
                {
                    "build_date": today_date_str,
                    "build_spec_hash": build_spec_hash,
                    "build_definition": build_definition,
                },
            )

            with log_file_lock, log_file.open(mode="a") as afd:
                afd.write(f"{stable_now_date().isoformat()}: {build_msg}...DONE\n")

        if filelock:
            filelock.lock.release()

        if dest_dir:
            copy_dir_tree(cached_repo_path, dest_dir)
            return dest_dir

        return cached_repo_path

    return _build_repo_w_cache_checking


@pytest.fixture(scope="session")
def teardown_cached_dir() -> Generator[TeardownCachedDirFn, None, None]:
    directories: list[Path] = []

    def _teardown_cached_dir(directory: Path | str) -> Path:
        directories.append(Path(directory))
        return directories[-1]

    try:
        yield _teardown_cached_dir
    finally:
        # clean up any registered cached directories
        for directory in directories:
            if directory.exists():
                remove_dir_tree(directory, force=True)


@pytest.fixture(scope="session")
def make_commit_obj(
    commit_author: Actor, stable_now_date: GetStableDateNowFn
) -> MakeCommitObjFn:
    def _make_commit(message: str) -> Commit:
        commit_timestamp = round(stable_now_date().timestamp())
        return Commit(
            repo=Repo(),
            binsha=Commit.NULL_BIN_SHA,
            message=message,
            author=commit_author,
            authored_date=commit_timestamp,
            committer=commit_author,
            committed_date=commit_timestamp,
            parents=[],
        )

    return _make_commit


@pytest.fixture(scope="session")
def get_md5_for_file() -> GetMd5ForFileFn:
    in_memory_cache = {}

    def _get_md5_for_file(file_path: Path | str) -> str:
        file_path = Path(file_path)
        rel_file_path = str(file_path.relative_to(PROJ_DIR))

        if rel_file_path not in in_memory_cache:
            in_memory_cache[rel_file_path] = md5(  # noqa: S324, not using hash for security
                file_path.read_bytes()
            ).hexdigest()

        return in_memory_cache[rel_file_path]

    return _get_md5_for_file


@pytest.fixture(scope="session")
def get_md5_for_set_of_files(
    get_md5_for_file: GetMd5ForFileFn,
) -> GetMd5ForSetOfFilesFn:
    in_memory_cache = {}

    def _get_md5_for_set_of_files(files: Sequence[Path | str]) -> str:
        # cast to a filtered and unique set of Path objects
        file_dependencies = sorted(
            set(
                filter(
                    lambda file_path: file_path.name != "__init__.py"
                    and file_path.stat().st_size > 0,
                    (Path(f).absolute().resolve() for f in files),
                )
            )
        )

        # create a hashable key of all dependencies to store the combined files hash
        cache_key = tuple(
            [str(file.relative_to(PROJ_DIR)) for file in file_dependencies]
        )

        # check if we have done this before
        if cache_key not in in_memory_cache:
            # since we haven't done this before, generate the hash for each file
            file_hashes = [get_md5_for_file(file) for file in file_dependencies]

            # combine the hashes into a string and then hash the result and store it
            in_memory_cache[cache_key] = md5(  # noqa: S324, not using hash for security
                str.join("\n", file_hashes).encode()
            ).hexdigest()

        # return the stored calculated hash for the set
        return in_memory_cache[cache_key]

    return _get_md5_for_set_of_files


@pytest.fixture(scope="session")
def clean_os_environment() -> dict[str, str]:
    return dict(
        filter(
            lambda k_v: k_v[1] is not None,  # type: ignore[arg-type]
            {
                "PATH": os.getenv("PATH"),
                "HOME": os.getenv("HOME"),
                **(
                    {}
                    if sys.platform != "win32"
                    else {
                        # Windows Required variables
                        "ALLUSERSAPPDATA": os.getenv("ALLUSERSAPPDATA"),
                        "ALLUSERSPROFILE": os.getenv("ALLUSERSPROFILE"),
                        "APPDATA": os.getenv("APPDATA"),
                        "COMMONPROGRAMFILES": os.getenv("COMMONPROGRAMFILES"),
                        "COMMONPROGRAMFILES(X86)": os.getenv("COMMONPROGRAMFILES(X86)"),
                        "DEFAULTUSERPROFILE": os.getenv("DEFAULTUSERPROFILE"),
                        "HOMEPATH": os.getenv("HOMEPATH"),
                        "PATHEXT": os.getenv("PATHEXT"),
                        "PROFILESFOLDER": os.getenv("PROFILESFOLDER"),
                        "PROGRAMFILES": os.getenv("PROGRAMFILES"),
                        "PROGRAMFILES(X86)": os.getenv("PROGRAMFILES(X86)"),
                        "SYSTEM": os.getenv("SYSTEM"),
                        "SYSTEM16": os.getenv("SYSTEM16"),
                        "SYSTEM32": os.getenv("SYSTEM32"),
                        "SYSTEMDRIVE": os.getenv("SYSTEMDRIVE"),
                        "SYSTEMROOT": os.getenv("SYSTEMROOT"),
                        "TEMP": os.getenv("TEMP"),
                        "TMP": os.getenv("TMP"),
                        "USERPROFILE": os.getenv("USERPROFILE"),
                        "USERSID": os.getenv("USERSID"),
                        "USERNAME": os.getenv("USERNAME"),
                        "WINDIR": os.getenv("WINDIR"),
                    }
                ),
            }.items(),
        )
    )



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple, NoReturn, TypeVar, Union

from semantic_release.commit_parser.util import force_str
from semantic_release.errors import CommitParseError

if TYPE_CHECKING:  # pragma: no cover
    from git.objects.commit import Commit

    from semantic_release.enums import LevelBump


class ParsedMessageResult(NamedTuple):
    """
    A read-only named tuple object representing the result from parsing a commit message.

    Essentially this is a data structure which holds the parsed information from a commit message
    without the actual commit object itself. Very helpful for unit testing.

    Most of the fields will replicate the fields of a
    :py:class:`ParsedCommit <semantic_release.commit_parser.token.ParsedCommit>`
    """

    bump: LevelBump
    type: str
    category: str
    scope: str
    descriptions: tuple[str, ...]
    breaking_descriptions: tuple[str, ...] = ()
    release_notices: tuple[str, ...] = ()
    linked_issues: tuple[str, ...] = ()
    linked_merge_request: str = ""
    include_in_changelog: bool = True


class ParsedCommit(NamedTuple):
    """A read-only named tuple object representing the result of parsing a commit message."""

    bump: LevelBump
    """A LevelBump enum value indicating what type of change this commit introduces."""

    type: str
    """
    The type of the commit as a string, per the commit message style.

    This is up to the parser to implement; for example, the EmojiCommitParser
    parser fills this field with the emoji representing the most significant change for
    the commit.
    """

    scope: str
    """
    The scope, as a string, parsed from the commit.

    Generally an optional field based on the commit message style which means it very likely can be an empty string.
    Commit styles which do not have a meaningful concept of "scope" usually fill this field with an empty string.
    """

    descriptions: list[str]
    """
    A list of paragraphs from the commit message.

    Paragraphs are generally delimited by a double-newline since git commit messages are sometimes manually wordwrapped with
    a single newline, but this is up to the parser to implement.
    """

    breaking_descriptions: list[str]
    """
    A list of paragraphs which are deemed to identify and describe breaking changes by the parser.

    An example would be a paragraph which begins with the text ``BREAKING CHANGE:`` in the commit message but
    the parser gennerally strips the prefix and includes the rest of the paragraph in this list.
    """

    commit: Commit
    """The original commit object (a class defined by GitPython) that was parsed"""

    release_notices: tuple[str, ...] = ()
    """
    A tuple of release notices, which are additional information about the changes that affect the user.

    An example would be a paragraph which begins with the text ``NOTICE:`` in the commit message but
    the parser generally strips the prefix and includes the rest of the paragraph in this list.
    """

    linked_issues: tuple[str, ...] = ()
    """
    A tuple of issue numbers as strings, if the commit is contains issue references.

    If there are no issue references, this should be an empty tuple. Although, we generally
    refer to them as "issue numbers", it generally should be a string to adhere to the prefixes
    used by the VCS (ex. ``#`` for GitHub, GitLab, etc.) or issue tracker (ex. JIRA uses ``AAA-###``).
    """

    linked_merge_request: str = ""
    """
    A pull request or merge request definition, if the commit is labeled with a pull/merge request number.

    This is a string value which includes any special character prefix used by the VCS
    (e.g. ``#`` for GitHub, ``!`` for GitLab).
    """

    include_in_changelog: bool = True
    """
    A boolean value indicating whether this commit should be included in the changelog.

    This enables parsers to flag commits which are not user-facing or are otherwise not
    relevant to the changelog to be filtered out by PSR's internal algorithms.
    """

    @property
    def message(self) -> str:
        """
        A string representation of the commit message.

        This is a pass through property for convience to access the ``message``
        attribute of the ``commit`` object.

        If the message is of type ``bytes`` then it will be decoded to a ``UTF-8`` string.
        """
        return force_str(self.commit.message).replace("\r", "")

    @property
    def hexsha(self) -> str:
        """
        A hex representation of the hash value of the commit.

        This is a pass through property for convience to access the ``hexsha`` attribute of the ``commit``.
        """
        return self.commit.hexsha

    @property
    def short_hash(self) -> str:
        """A short representation of the hash value (in hex) of the commit."""
        return self.hexsha[:7]

    @property
    def linked_pull_request(self) -> str:
        """An alias to the linked_merge_request attribute."""
        return self.linked_merge_request

    def is_merge_commit(self) -> bool:
        return bool(len(self.commit.parents) > 1)

    @staticmethod
    def from_parsed_message_result(
        commit: Commit, parsed_message_result: ParsedMessageResult
    ) -> ParsedCommit:
        """A convience method to create a ParsedCommit object from a ParsedMessageResult object and a Commit object."""
        return ParsedCommit(
            bump=parsed_message_result.bump,
            # TODO: breaking v11, swap back to type rather than category
            type=parsed_message_result.category,
            scope=parsed_message_result.scope,
            descriptions=list(parsed_message_result.descriptions),
            breaking_descriptions=list(parsed_message_result.breaking_descriptions),
            commit=commit,
            release_notices=parsed_message_result.release_notices,
            linked_issues=parsed_message_result.linked_issues,
            linked_merge_request=parsed_message_result.linked_merge_request,
            include_in_changelog=parsed_message_result.include_in_changelog,
        )


class ParseError(NamedTuple):
    """A read-only named tuple object representing an error that occurred while parsing a commit message."""

    commit: Commit
    """The original commit object (a class defined by GitPython) that was parsed"""

    error: str
    """A string with a description for why the commit parsing failed."""

    @property
    def message(self) -> str:
        """
        A string representation of the commit message.

        This is a pass through property for convience to access the ``message``
        attribute of the ``commit`` object.

        If the message is of type ``bytes`` then it will be decoded to a ``UTF-8`` string.
        """
        return force_str(self.commit.message).replace("\r", "")

    @property
    def hexsha(self) -> str:
        """
        A hex representation of the hash value of the commit.

        This is a pass through property for convience to access the ``hexsha`` attribute of the ``commit``.
        """
        return self.commit.hexsha

    @property
    def short_hash(self) -> str:
        """A short representation of the hash value (in hex) of the commit."""
        return self.hexsha[:7]

    def is_merge_commit(self) -> bool:
        return bool(len(self.commit.parents) > 1)

    def raise_error(self) -> NoReturn:
        """A convience method to raise a CommitParseError with the error message."""
        raise CommitParseError(self.error)


_T = TypeVar("_T", bound=ParsedCommit)
_E = TypeVar("_E", bound=ParseError)

# For extensions, this type can be used to build an alias
# for example CustomParseResult = ParseResultType[CustomParsedCommit, ParseError]
ParseResultType = Union[_T, _E]
ParseResult = ParseResultType[ParsedCommit, ParseError]



from __future__ import annotations

import re
from collections import defaultdict
from logging import Filter as LoggingFilter
from typing import TYPE_CHECKING

from semantic_release.globals import logger

if TYPE_CHECKING:  # pragma: no cover
    from logging import LogRecord
    from typing import Iterable


# https://relaxdiego.com/2014/07/logging-in-python.html
# Updated/adapted for Python3
class MaskingFilter(LoggingFilter):
    REPLACE_STR = "*" * 4
    _UNWANTED = frozenset([s for obj in ("", None) for s in (repr(obj), str(obj))])

    def __init__(
        self,
        _use_named_masks: bool = False,
        **patterns: Iterable[str | re.Pattern[str]],
    ) -> None:
        super().__init__()
        self._redact_patterns = defaultdict(set)
        for k, vs in patterns.items():
            self._redact_patterns[k] = {v for v in vs if v and v not in self._UNWANTED}
        self._use_named_masks = _use_named_masks

    def add_mask_for(self, data: str, name: str = "redacted") -> MaskingFilter:
        if data and data not in self._UNWANTED:
            logger.debug("Adding redact pattern '%r' to redact_patterns", name)
            self._redact_patterns[name].add(data)
        return self

    def filter(self, record: LogRecord) -> bool:
        # Note if we blindly mask all types, we will actually cast arguments to
        # log functions from external libraries to strings before they are
        # formatted into the message - for example, a dependency calling
        # log.debug("%d", 15) will raise a TypeError as this filter would
        # otherwise convert 15 to "15", and "%d" % "15" raises the error.
        # One may find a specific example of where this issue could manifest itself
        # here: https://github.com/urllib3/urllib3/blob/a5b29ac1025f9bb30f2c9b756f3b171389c2c039/src/urllib3/connectionpool.py#L1003
        # Anything which could reasonably be expected to be logged without being
        # cast to a string should be excluded from the cast here.
        record.msg = self.mask(record.msg)
        if record.args is None:
            pass
        elif isinstance(record.args, dict):
            record.args = {
                k: v if type(v) in (bool, int, float) else self.mask(str(v))
                for k, v in record.args.items()
            }
        else:
            record.args = tuple(
                arg if type(arg) in (bool, int, float) else self.mask(str(arg))
                for arg in record.args
            )
        return True

    def mask(self, msg: str) -> str:
        if not isinstance(msg, str):
            logger.debug("cannot mask object of type %s", type(msg))
            return msg
        for mask, values in self._redact_patterns.items():
            repl_string = (
                self.REPLACE_STR
                if not self._use_named_masks
                else f"<{mask!r} (value removed)>"
            )
            for data in values:
                if isinstance(data, str):
                    msg = msg.replace(data, repl_string)
                elif isinstance(data, re.Pattern):
                    msg = data.sub(repl_string, msg)
        return msg



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from git import Repo

from semantic_release.hvcs.github import Github

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from typing import Protocol, Sequence

    from click.testing import Result

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        RepoActionConfigure,
        RepoActionConfigureMonorepo,
        RepoActionCreateMonorepo,
    )

    class InitMirrorRepo4RebuildFn(Protocol):
        def __call__(
            self,
            mirror_repo_dir: Path,
            configuration_steps: Sequence[
                RepoActionConfigure
                | RepoActionCreateMonorepo
                | RepoActionConfigureMonorepo
            ],
            files_to_remove: Sequence[Path],
        ) -> Path: ...

    class RunPSReleaseFn(Protocol):
        def __call__(
            self,
            next_version_str: str,
            git_repo: Repo,
            config_toml_path: Path = ...,
        ) -> Result: ...


@pytest.fixture(scope="session")
def init_mirror_repo_for_rebuild(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
) -> InitMirrorRepo4RebuildFn:
    def _init_mirror_repo_for_rebuild(
        mirror_repo_dir: Path,
        configuration_steps: Sequence[
            RepoActionConfigure | RepoActionCreateMonorepo | RepoActionConfigureMonorepo
        ],
        files_to_remove: Sequence[Path],
    ) -> Path:
        # Create the mirror repo directory
        mirror_repo_dir.mkdir(exist_ok=True, parents=True)

        # Initialize mirror repository
        build_repo_from_definition(
            dest_dir=mirror_repo_dir,
            repo_construction_steps=configuration_steps,
        )

        with Repo(mirror_repo_dir) as mirror_git_repo:
            for filepath in files_to_remove:
                file = (
                    (mirror_git_repo.working_dir / filepath).resolve().absolute()
                    if not filepath.is_absolute()
                    else filepath
                )
                if (
                    Path(mirror_git_repo.working_dir) not in file.parents
                    or not file.exists()
                ):
                    continue

                mirror_git_repo.git.rm(str(file), force=True)

        return mirror_repo_dir

    return _init_mirror_repo_for_rebuild


@pytest.fixture(scope="session")
def run_psr_release(
    run_cli: RunCliFn,
    changelog_rst_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_file: Path,
) -> RunPSReleaseFn:
    base_version_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
    write_changelog_only_cmd = [
        *base_version_cmd,
        "--changelog",
        "--no-commit",
        "--no-tag",
        "--skip-build",
    ]

    def _run_psr_release(
        next_version_str: str,
        git_repo: Repo,
        config_toml_path: Path = pyproject_toml_file,
    ) -> Result:
        version_n_buildmeta = next_version_str.split("+", maxsplit=1)
        version_n_prerelease = version_n_buildmeta[0].split("-", maxsplit=1)

        build_metadata_args = (
            ["--build-metadata", version_n_buildmeta[-1]]
            if len(version_n_buildmeta) > 1
            else []
        )

        prerelease_args = (
            [
                "--as-prerelease",
                "--prerelease-token",
                version_n_prerelease[-1].split(".", maxsplit=1)[0],
            ]
            if len(version_n_prerelease) > 1
            else []
        )

        # Initial run to write the RST changelog
        # 1. configure PSR to write the RST changelog with the RST default insertion flag
        update_pyproject_toml(
            "tool.semantic_release.changelog.default_templates.changelog_file",
            str(changelog_rst_file),
            toml_file=config_toml_path,
        )
        cli_cmd = [*write_changelog_only_cmd, *prerelease_args, *build_metadata_args]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})
        assert_successful_exit_code(result, cli_cmd)

        # Reset the index in case PSR added anything to the index
        git_repo.git.reset("--mixed", "HEAD")

        # Add the changelog file to the git index but reset the working directory
        git_repo.git.add(str(changelog_rst_file.resolve()))
        git_repo.git.checkout("--", ".")

        # Actual run to release & write the MD changelog
        cli_cmd = [
            *base_version_cmd,
            *prerelease_args,
            *build_metadata_args,
        ]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})
        assert_successful_exit_code(result, cli_cmd)

        return result

    return _run_psr_release



from __future__ import annotations

import os
from unittest import mock

import pytest

from semantic_release.hvcs.bitbucket import Bitbucket

from tests.const import EXAMPLE_HVCS_DOMAIN, EXAMPLE_REPO_NAME, EXAMPLE_REPO_OWNER


@pytest.fixture
def default_bitbucket_client():
    remote_url = (
        f"git@{Bitbucket.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git"
    )
    return Bitbucket(remote_url=remote_url)


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "patched_os_environ",
            "hvcs_domain",
            "hvcs_api_domain",
            "expected_hvcs_domain",
            "expected_api_url",
            "insecure",
        ],
    ),
    [
        # No env vars as CI is handled by Bamboo or Jenkins (which require user defined defaults)
        # API paths are different in BitBucket Cloud (bitbucket.org) vs BitBucket Data Center
        (
            # Default values (BitBucket Cloud)
            {},
            None,
            None,
            f"https://{Bitbucket.DEFAULT_DOMAIN}",
            Bitbucket.DEFAULT_API_URL_CLOUD,
            False,
        ),
        (
            # Explicitly set default values
            {},
            Bitbucket.DEFAULT_DOMAIN,
            Bitbucket.DEFAULT_API_URL_CLOUD,
            f"https://{Bitbucket.DEFAULT_DOMAIN}",
            Bitbucket.DEFAULT_API_URL_CLOUD,
            False,
        ),
        (
            # Explicitly set custom values with full api path
            {},
            EXAMPLE_HVCS_DOMAIN,
            f"{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            False,
        ),
        (
            # Explicitly defined api as subdomain
            # POSSIBLY WRONG ASSUMPTION of Api path for BitBucket Server
            {},
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://api.{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://api.{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            False,
        ),
        (
            # Custom domain for on premise BitBucket Server (derive api endpoint)
            {},
            EXAMPLE_HVCS_DOMAIN,
            None,
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            False,
        ),
        (
            # Custom domain with path prefix
            {},
            "special.custom.server/bitbucket",
            None,
            "https://special.custom.server/bitbucket",
            "https://special.custom.server/bitbucket/rest/api/1.0",
            False,
        ),
        (
            # Allow insecure http connections explicitly
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            True,
        ),
        (
            # Allow insecure http connections explicitly & imply insecure api domain
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            None,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            True,
        ),
        (
            # Infer insecure connection from user configuration
            {},
            EXAMPLE_HVCS_DOMAIN,
            f"{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            True,
        ),
        (
            # Infer insecure connection from user configuration & imply insecure api domain
            {},
            EXAMPLE_HVCS_DOMAIN,
            None,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/rest/api/1.0",
            True,
        ),
    ],
)
@pytest.mark.parametrize(
    "remote_url",
    [
        f"git@{Bitbucket.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
        f"https://{Bitbucket.DEFAULT_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
    ],
)
@pytest.mark.parametrize("token", ("abc123", None))
def test_bitbucket_client_init(
    patched_os_environ: dict[str, str],
    hvcs_domain: str | None,
    hvcs_api_domain: str | None,
    expected_hvcs_domain: str,
    expected_api_url: str,
    remote_url: str,
    token: str | None,
    insecure: bool,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        client = Bitbucket(
            remote_url=remote_url,
            hvcs_domain=hvcs_domain,
            hvcs_api_domain=hvcs_api_domain,
            token=token,
            allow_insecure=insecure,
        )

        assert expected_hvcs_domain == str(client.hvcs_domain)
        assert expected_api_url == str(client.api_url)
        assert token == client.token
        assert remote_url == client._remote_url


@pytest.mark.parametrize(
    "hvcs_domain, hvcs_api_domain, insecure",
    [
        # Bad base domain schemes
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", None, False),
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", None, True),
        # Unallowed insecure connections when base domain is insecure
        (f"http://{EXAMPLE_HVCS_DOMAIN}", None, False),
        # Bad API domain schemes
        (None, f"ftp://api.{EXAMPLE_HVCS_DOMAIN}", False),
        (None, f"ftp://api.{EXAMPLE_HVCS_DOMAIN}", True),
        # Unallowed insecure connections when api domain is insecure
        (None, f"http://{EXAMPLE_HVCS_DOMAIN}", False),
    ],
)
def test_bitbucket_client_init_with_invalid_scheme(
    hvcs_domain: str | None,
    hvcs_api_domain: str | None,
    insecure: bool,
):
    with pytest.raises(ValueError), mock.patch.dict(os.environ, {}, clear=True):
        Bitbucket(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=hvcs_domain,
            hvcs_api_domain=hvcs_api_domain,
            allow_insecure=insecure,
        )


@pytest.mark.parametrize(
    "patched_os_environ, expected_owner, expected_name",
    [
        ({}, None, None),
        ({"BITBUCKET_REPO_FULL_NAME": "path/to/repo/foo"}, "path/to/repo", "foo"),
    ],
)
def test_bitbucket_get_repository_owner_and_name(
    default_bitbucket_client: Bitbucket,
    patched_os_environ: dict[str, str],
    expected_owner: str,
    expected_name: str,
):
    # expected results should be a tuple[namespace, repo_name]
    # when None, the default values are used which matches default_bitbucket_client's setup
    expected_result = (
        expected_owner or EXAMPLE_REPO_OWNER,
        expected_name or EXAMPLE_REPO_NAME,
    )

    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        # Execute in mocked environment
        result = default_bitbucket_client._get_repository_owner_and_name()

        # Evaluate (expected -> actual)
        assert expected_result == result


def test_compare_url(default_bitbucket_client: Bitbucket):
    start_rev = "revA"
    end_rev = "revB"
    expected_url = (
        "{server}/{owner}/{repo}/branches/compare/{from_rev}%0D{to_rev}".format(
            server=default_bitbucket_client.hvcs_domain.url,
            owner=default_bitbucket_client.owner,
            repo=default_bitbucket_client.repo_name,
            from_rev=start_rev,
            to_rev=end_rev,
        )
    )
    actual_url = default_bitbucket_client.compare_url(
        from_rev=start_rev, to_rev=end_rev
    )
    assert expected_url == actual_url


@pytest.mark.parametrize(
    "patched_os_environ, use_token, token, remote_url, expected_auth_url",
    [
        (
            {"BITBUCKET_USER": "foo"},
            False,
            "",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {},
            False,
            "aabbcc",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {},
            True,
            "aabbcc",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
            f"https://x-token-auth:aabbcc@{Bitbucket.DEFAULT_DOMAIN}/custom/example.git",
        ),
        (
            {"BITBUCKET_USER": "foo"},
            False,
            "aabbcc",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {"BITBUCKET_USER": "foo"},
            True,
            "aabbcc",
            f"git@{Bitbucket.DEFAULT_DOMAIN}:custom/example.git",
            f"https://foo:aabbcc@{Bitbucket.DEFAULT_DOMAIN}/custom/example.git",
        ),
    ],
)
def test_remote_url(
    default_bitbucket_client: Bitbucket,
    patched_os_environ: dict[str, str],
    use_token: bool,
    token: str,
    remote_url: str,
    expected_auth_url: str,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        default_bitbucket_client._remote_url = remote_url
        default_bitbucket_client.token = token
        assert expected_auth_url == default_bitbucket_client.remote_url(
            use_token=use_token
        )


def test_commit_hash_url(default_bitbucket_client: Bitbucket):
    sha = "244f7e11bcb1e1ce097db61594056bc2a32189a0"
    expected_url = "{server}/{owner}/{repo}/commits/{sha}".format(
        server=default_bitbucket_client.hvcs_domain,
        owner=default_bitbucket_client.owner,
        repo=default_bitbucket_client.repo_name,
        sha=sha,
    )
    assert expected_url == default_bitbucket_client.commit_hash_url(sha)


def test_commit_hash_url_w_custom_server():
    """
    Test the commit hash URL generation for a self-hosted Bitbucket server with prefix.

    ref: https://github.com/python-semantic-release/python-semantic-release/issues/1204
    """
    sha = "244f7e11bcb1e1ce097db61594056bc2a32189a0"
    expected_url = "{server}/{owner}/{repo}/commits/{sha}".format(
        server=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        owner="foo",
        repo=EXAMPLE_REPO_NAME,
        sha=sha,
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        actual_url = Bitbucket(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo/foo/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        ).commit_hash_url(sha)

    assert expected_url == actual_url


@pytest.mark.parametrize("pr_number", (666, "666", "#666"))
def test_pull_request_url(default_bitbucket_client: Bitbucket, pr_number: int | str):
    expected_url = "{server}/{owner}/{repo}/pull-requests/{pr_number}".format(
        server=default_bitbucket_client.hvcs_domain,
        owner=default_bitbucket_client.owner,
        repo=default_bitbucket_client.repo_name,
        pr_number=str(pr_number).lstrip("#"),
    )
    actual_url = default_bitbucket_client.pull_request_url(pr_number=pr_number)
    assert expected_url == actual_url



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from semantic_release.commit_parser.token import ParseResultType

if TYPE_CHECKING:  # pragma: no cover
    from git.objects.commit import Commit


class ParserOptions(dict):
    """
    ParserOptions should accept the keyword arguments they are interested in
    from configuration and process them as desired, ultimately creating attributes
    on an instance which can be accessed by the corresponding commit parser.

    For example:
    >>> class MyParserOptions(ParserOptions):
    ...     def __init__(self, message_prefix: str) -> None:
    ...         self.prefix = message_prefix * 2

    >>> class MyCommitParser(AbstractCommitParser):
    ...     parser_options = MyParserOptions
    ...
    ...     def parse(self, Commit):
    ...         print(self.options.prefix)
    ...         ...

    Any defaults that need to be set should also be done in this class too.
    Invalid options should be signalled by raising an ``InvalidOptionsException``
    within the ``__init__`` method of the options class.

    A dataclass is also well suited to this; if type-checking of input is desired,
    a ``pydantic.dataclasses.dataclass`` works well and is used internally
    by python-semantic-release. Parser options are not validated in the configuration
    and passed directly to the appropriate class to handle.
    """

    def __init__(self, **_: Any) -> None:
        pass


# TT = TokenType, a subclass of ParsedCommit
_TT = TypeVar("_TT", bound=ParseResultType)
_OPTS = TypeVar("_OPTS", bound=ParserOptions)


class CommitParser(ABC, Generic[_TT, _OPTS]):
    """
    Abstract base class for all commit parsers. Custom commit parsers should inherit
    from this class.

    A class-level ``parser_options`` attribute should be set to a subclass of
    ``BaseParserOptions``; this will be used to provide the default options to the
    parser. Note that a nested class can be used directly, if preferred:

    >>> class MyParser(CommitParser):
            @dataclass
            class parser_options(ParserOptions):
                allowed_types: Tuple[str] = ("feat", "fix", "docs")
                major_types: Tuple[str] = ("breaking",)
                minor_types: Tuple[str] = ("fix", "patch")
                ...
            def __init__(self, options: parser_options) -> None:
                ...
    """

    # TODO: Deprecate in lieu of get_default_options()
    parser_options: type[ParserOptions] = ParserOptions

    def __init__(self, options: _OPTS | None = None) -> None:
        self.options: _OPTS = (
            options if options is not None else self.get_default_options()
        )

    # TODO: BREAKING CHANGE v11, add abstract method for all custom parsers
    # @staticmethod
    # @abstractmethod
    def get_default_options(self) -> _OPTS:
        return self.parser_options()  # type: ignore[return-value]

    @abstractmethod
    def parse(self, commit: Commit) -> _TT | list[_TT]: ...



from __future__ import annotations

from datetime import timedelta
from itertools import count
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest

from semantic_release.cli.config import ChangelogOutputFormat
from semantic_release.version.version import Version

import tests.conftest
import tests.const
import tests.util
from tests.const import (
    DEFAULT_BRANCH_NAME,
    EXAMPLE_HVCS_DOMAIN,
    INITIAL_COMMIT_MESSAGE,
    RepoActionStep,
)

if TYPE_CHECKING:
    from typing import Any, Generator, Sequence

    from semantic_release.commit_parser._base import CommitParser, ParserOptions
    from semantic_release.commit_parser.conventional.parser import (
        ConventionalCommitParser,
    )
    from semantic_release.commit_parser.emoji import EmojiCommitParser
    from semantic_release.commit_parser.scipy import ScipyCommitParser
    from semantic_release.commit_parser.token import ParseResult

    from tests.conftest import (
        GetCachedRepoDataFn,
        GetMd5ForSetOfFilesFn,
        GetStableDateNowFn,
    )
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuildRepoOrCopyCacheFn,
        BuildSpecificRepoFn,
        BuiltRepoResult,
        CommitConvention,
        ConvertCommitSpecsToCommitDefsFn,
        ConvertCommitSpecToCommitDefFn,
        ExProjectGitRepoFn,
        FormatGitMergeCommitMsgFn,
        GetRepoDefinitionFn,
        RepoActionGitFFMergeDetails,
        RepoActionGitMerge,
        RepoActionGitMergeDetails,
        RepoActions,
        RepoActionWriteChangelogsDestFile,
        TomlSerializableTypes,
    )


BETA_BRANCH_NAME = "beta"
DEV_BRANCH_NAME = "dev"
FEAT_BRANCH_1_NAME = "feat/feature-1"
FEAT_BRANCH_2_NAME = "feat/feature-2"
FEAT_BRANCH_3_NAME = "feat/feature-3"
FEAT_BRANCH_4_NAME = "feat/feature-4"
FIX_BRANCH_1_NAME = "fix/patch-1"
FIX_BRANCH_2_NAME = "fix/patch-2"


@pytest.fixture(scope="session")
def deps_files_4_git_flow_repo_w_4_release_channels(
    deps_files_4_example_git_project: list[Path],
) -> list[Path]:
    return [
        *deps_files_4_example_git_project,
        # This file
        Path(__file__).absolute(),
        # because of imports
        Path(tests.const.__file__).absolute(),
        Path(tests.util.__file__).absolute(),
        # because of the fixtures
        Path(tests.conftest.__file__).absolute(),
    ]


@pytest.fixture(scope="session")
def build_spec_hash_4_git_flow_repo_w_4_release_channels(
    get_md5_for_set_of_files: GetMd5ForSetOfFilesFn,
    deps_files_4_git_flow_repo_w_4_release_channels: list[Path],
) -> str:
    # Generates a hash of the build spec to set when to invalidate the cache
    return get_md5_for_set_of_files(deps_files_4_git_flow_repo_w_4_release_channels)


@pytest.fixture(scope="session")
def get_repo_definition_4_git_flow_repo_w_4_release_channels(
    convert_commit_specs_to_commit_defs: ConvertCommitSpecsToCommitDefsFn,
    convert_commit_spec_to_commit_def: ConvertCommitSpecToCommitDefFn,
    format_merge_commit_msg_git: FormatGitMergeCommitMsgFn,
    changelog_md_file: Path,
    changelog_rst_file: Path,
    stable_now_date: GetStableDateNowFn,
    default_conventional_parser: ConventionalCommitParser,
    default_emoji_parser: EmojiCommitParser,
    default_scipy_parser: ScipyCommitParser,
    default_tag_format_str: str,
) -> GetRepoDefinitionFn:
    """
    This fixture returns a function that when called will define the actions needed to
    build a git repo that uses the git flow branching strategy and git merge commits
    with 4 release channels.

    This very complex repository mirrors the git flow example provided by a user in
    issue [#789](https://github.com/python-semantic-release/python-semantic-release/issues/789).

        1. [feature branches] revision releases which include build-metadata of the branch
            name (slightly differs from user where the release also used alpha+build-metadata)

        2. [dev branch] alpha feature releases (x.x.x-alpha.x)

        3. [beta branch] beta releases (x.x.x-beta.x)

        4. [main branch] official (production) releases (x.x.x)
    """
    parser_classes: dict[CommitConvention, CommitParser[Any, Any]] = {
        "conventional": default_conventional_parser,
        "emoji": default_emoji_parser,
        "scipy": default_scipy_parser,
    }

    def _get_repo_from_definition(
        commit_type: CommitConvention,
        hvcs_client_name: str = "github",
        hvcs_domain: str = EXAMPLE_HVCS_DOMAIN,
        tag_format_str: str | None = None,
        extra_configs: dict[str, TomlSerializableTypes] | None = None,
        mask_initial_release: bool = True,
        ignore_merge_commits: bool = True,
    ) -> Sequence[RepoActions]:
        stable_now_datetime = stable_now_date()
        commit_timestamp_gen = (
            (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
            for i in count(step=1)
        )
        commit_parser = cast(
            "CommitParser[ParseResult, ParserOptions]",
            parser_classes[commit_type],
        )

        # Common static actions or components
        changelog_file_definitions: Sequence[RepoActionWriteChangelogsDestFile] = [
            {
                "path": changelog_md_file,
                "format": ChangelogOutputFormat.MARKDOWN,
                "mask_initial_release": mask_initial_release,
            },
            {
                "path": changelog_rst_file,
                "format": ChangelogOutputFormat.RESTRUCTURED_TEXT,
                "mask_initial_release": mask_initial_release,
            },
        ]

        ff_beta_branch_merge_def: RepoActionGitMerge[RepoActionGitFFMergeDetails] = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": BETA_BRANCH_NAME,
                "fast_forward": True,
            },
        }

        fast_forward_dev_branch_actions: Sequence[RepoActions] = [
            {
                "action": RepoActionStep.GIT_CHECKOUT,
                "details": {"branch": DEV_BRANCH_NAME},
            },
            ff_beta_branch_merge_def,
        ]

        ff_main_branch_merge_def: RepoActionGitMerge[RepoActionGitFFMergeDetails] = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": DEFAULT_BRANCH_NAME,
                "fast_forward": True,
            },
        }

        fast_forward_beta_branch_actions: Sequence[RepoActions] = [
            {
                "action": RepoActionStep.GIT_CHECKOUT,
                "details": {"branch": BETA_BRANCH_NAME},
            },
            ff_main_branch_merge_def,
        ]

        merge_dev_into_beta_gen: Generator[
            RepoActionGitMerge[RepoActionGitMergeDetails], None, None
        ] = (
            {
                "action": RepoActionStep.GIT_MERGE,
                "details": {
                    "branch_name": DEV_BRANCH_NAME,
                    "fast_forward": False,
                    "commit_def": convert_commit_spec_to_commit_def(
                        {
                            "cid": f"merge-dev2beta-{i}",
                            "conventional": (
                                merge_msg := format_merge_commit_msg_git(
                                    branch_name=DEV_BRANCH_NAME,
                                    tgt_branch_name=BETA_BRANCH_NAME,
                                )
                            ),
                            "emoji": merge_msg,
                            "scipy": merge_msg,
                            "datetime": next(commit_timestamp_gen),
                            "include_in_changelog": not ignore_merge_commits,
                        },
                        commit_type,
                        parser=commit_parser,
                    ),
                },
            }
            for i in count(start=1)
        )

        merge_beta_into_main_gen: Generator[
            RepoActionGitMerge[RepoActionGitMergeDetails], None, None
        ] = (
            {
                "action": RepoActionStep.GIT_MERGE,
                "details": {
                    "branch_name": BETA_BRANCH_NAME,
                    "fast_forward": False,
                    "commit_def": convert_commit_spec_to_commit_def(
                        {
                            "cid": f"merge-beta2main-{i}",
                            "conventional": (
                                merge_msg := format_merge_commit_msg_git(
                                    branch_name=BETA_BRANCH_NAME,
                                    tgt_branch_name=DEFAULT_BRANCH_NAME,
                                )
                            ),
                            "emoji": merge_msg,
                            "scipy": merge_msg,
                            "datetime": next(commit_timestamp_gen),
                            "include_in_changelog": not ignore_merge_commits,
                        },
                        commit_type,
                        parser=commit_parser,
                    ),
                },
            }
            for i in count(start=1)
        )

        # Define All the steps required to create the repository
        repo_construction_steps: list[RepoActions] = []

        repo_construction_steps.append(
            {
                "action": RepoActionStep.CONFIGURE,
                "details": {
                    "commit_type": commit_type,
                    "hvcs_client_name": hvcs_client_name,
                    "hvcs_domain": hvcs_domain,
                    "tag_format_str": tag_format_str or default_tag_format_str,
                    "mask_initial_release": mask_initial_release,
                    "extra_configs": {
                        # Set the default release branch
                        "tool.semantic_release.branches.main": {
                            "match": rf"^{DEFAULT_BRANCH_NAME}$",
                            "prerelease": False,
                        },
                        # branch "beta" has prerelease suffix of "beta"
                        "tool.semantic_release.branches.beta": {
                            "match": rf"^{BETA_BRANCH_NAME}$",
                            "prerelease": True,
                            "prerelease_token": "beta",
                        },
                        # branch "development" has prerelease suffix of "alpha"
                        "tool.semantic_release.branches.dev": {
                            "match": rf"^{DEV_BRANCH_NAME}$",
                            "prerelease": True,
                            "prerelease_token": "alpha",
                        },
                        # branch "feat/*" has prerelease suffix of "rev"
                        "tool.semantic_release.branches.features": {
                            "match": r"^feat/.+",
                            "prerelease": True,
                            "prerelease_token": "rev",
                        },
                        "tool.semantic_release.allow_zero_version": False,
                        "tool.semantic_release.commit_parser_options.ignore_merge_commits": ignore_merge_commits,
                        **(extra_configs or {}),
                    },
                },
            }
        )

        # Make initial release
        new_version = Version.parse(
            "1.0.0", tag_format=(tag_format_str or default_tag_format_str)
        )

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": [
                            # only one commit to start the main branch
                            convert_commit_spec_to_commit_def(
                                {
                                    "cid": (
                                        cid_db_c1_initial := "db_c1_initial_commit"
                                    ),
                                    "conventional": INITIAL_COMMIT_MESSAGE,
                                    "emoji": INITIAL_COMMIT_MESSAGE,
                                    "scipy": INITIAL_COMMIT_MESSAGE,
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": bool(
                                        commit_type == "emoji"
                                    ),
                                },
                                commit_type,
                                parser=commit_parser,
                            ),
                        ],
                    },
                },
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": BETA_BRANCH_NAME,
                            "start_branch": DEFAULT_BRANCH_NAME,
                        },
                    },
                },
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": DEV_BRANCH_NAME,
                            "start_branch": BETA_BRANCH_NAME,
                        },
                    },
                },
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": FEAT_BRANCH_1_NAME,
                            "start_branch": DEV_BRANCH_NAME,
                        },
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (cid_feb1_c1_feat := "feat_branch1_c1_feat"),
                                    "conventional": "feat: add new feature",
                                    "emoji": ":sparkles: add new feature",
                                    "scipy": "ENH: add new feature",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=commit_parser,
                        ),
                    },
                },
            ]
        )

        merge_def_type_placeholder: RepoActionGitMerge[RepoActionGitMergeDetails] = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": FEAT_BRANCH_1_NAME,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_feb1_merge2dev := "feat_branch1_merge2dev"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_git(
                                branch_name=FEAT_BRANCH_1_NAME,
                                tgt_branch_name=DEV_BRANCH_NAME,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=commit_parser,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEV_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": BETA_BRANCH_NAME},
                },
                {
                    **(merge_dev_into_beta_1 := next(merge_dev_into_beta_gen)),
                },
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEFAULT_BRANCH_NAME},
                },
                {
                    **(merge_beta_into_main_1 := next(merge_beta_into_main_gen)),
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        cid_db_c1_initial,
                                        cid_feb1_c1_feat,
                                        cid_feb1_merge2dev,
                                        merge_dev_into_beta_1["details"]["commit_def"][
                                            "cid"
                                        ],
                                        merge_beta_into_main_1["details"]["commit_def"][
                                            "cid"
                                        ],
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Make a fix and release it as an alpha release
        new_version = Version.parse("1.0.1-alpha.1", tag_format=new_version.tag_format)
        repo_construction_steps.extend(
            [
                *fast_forward_beta_branch_actions,
                *fast_forward_dev_branch_actions,
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": FIX_BRANCH_1_NAME,
                            "start_branch": DEV_BRANCH_NAME,
                        }
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (cid_fib1_c1_fix := "fix_branch_1_c1_fix"),
                                    "conventional": "fix(cli): fix config cli command\n\nCloses: #123\n",
                                    "emoji": ":bug: (cli) fix config cli command\n\nCloses: #123\n",
                                    "scipy": "BUG:cli: fix config cli command\n\nCloses: #123\n",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=commit_parser,
                        ),
                    },
                },
            ]
        )

        merge_def_type_placeholder = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": FIX_BRANCH_1_NAME,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_fib1_merge2dev := "fix_branch_1_merge_2_dev"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_git(
                                branch_name=FIX_BRANCH_1_NAME,
                                tgt_branch_name=DEV_BRANCH_NAME,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=commit_parser,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEV_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        cid_fib1_c1_fix,
                                        cid_fib1_merge2dev,
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Merge in the successful alpha release and create a beta release
        new_version = Version.parse("1.0.1-beta.1", tag_format=new_version.tag_format)
        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": BETA_BRANCH_NAME},
                },
                {
                    **(merge_dev_into_beta_2 := next(merge_dev_into_beta_gen)),
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        merge_dev_into_beta_2["details"]["commit_def"][
                                            "cid"
                                        ],
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Fix a bug found in beta release and create a new alpha release
        new_version = Version.parse("1.0.1-alpha.2", tag_format=new_version.tag_format)
        repo_construction_steps.extend(
            [
                *fast_forward_dev_branch_actions,
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": FIX_BRANCH_2_NAME,
                            "start_branch": DEV_BRANCH_NAME,
                        }
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (cid_fib2_c1_fix := "fix_branch_2_c1_fix"),
                                    "conventional": "fix(config): fix config option",
                                    "emoji": ":bug: (config) fix config option",
                                    "scipy": "BUG: config: fix config option",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=commit_parser,
                        ),
                    },
                },
            ]
        )

        merge_def_type_placeholder = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": FIX_BRANCH_2_NAME,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_fib2_merge2dev := "fix_branch_2_merge_2_dev"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_git(
                                branch_name=FIX_BRANCH_2_NAME,
                                tgt_branch_name=DEV_BRANCH_NAME,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=commit_parser,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEV_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        cid_fib2_c1_fix,
                                        cid_fib2_merge2dev,
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Merge in the 2nd successful alpha release and create a secondary beta release
        new_version = Version.parse("1.0.1-beta.2", tag_format=new_version.tag_format)

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": BETA_BRANCH_NAME},
                },
                {
                    **(merge_dev_into_beta_3 := next(merge_dev_into_beta_gen)),
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        merge_dev_into_beta_3["details"]["commit_def"][
                                            "cid"
                                        ],
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Add a new feature (another developer was working on) and create a release for it
        # Based on Semver standard, Build metadata is restricted to [A-Za-z0-9-] so we replace the '/' with a '-'
        new_version = Version.parse(
            f"""1.1.0-rev.1+{FEAT_BRANCH_2_NAME.replace("/", '-')}""",
            tag_format=new_version.tag_format,
        )

        repo_construction_steps.extend(
            [
                *fast_forward_dev_branch_actions,
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": FEAT_BRANCH_2_NAME,
                            "start_branch": DEV_BRANCH_NAME,
                        }
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (
                                        cid_feb2_c1_feat := "feat_branch_2_c1_feat"
                                    ),
                                    "conventional": "feat(feat-2): add another primary feature",
                                    "emoji": ":sparkles: (feat-2) add another primary feature",
                                    "scipy": "ENH: feat-2: add another primary feature",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=commit_parser,
                        ),
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        cid_feb2_c1_feat,
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Merge in the successful revision release and create an alpha release
        new_version = Version.parse("1.1.0-alpha.1", tag_format=new_version.tag_format)

        merge_def_type_placeholder = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": FEAT_BRANCH_2_NAME,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_feb2_merge2dev := "feat_branch_2_merge_2_dev"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_git(
                                branch_name=FEAT_BRANCH_2_NAME,
                                tgt_branch_name=DEV_BRANCH_NAME,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=commit_parser,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEV_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        cid_feb2_merge2dev,
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # Merge in the successful alpha release and create a beta release
        new_version = Version.parse("1.1.0-beta.1", tag_format=new_version.tag_format)

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": BETA_BRANCH_NAME},
                },
                {
                    **(merge_dev_into_beta_4 := next(merge_dev_into_beta_gen)),
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        merge_dev_into_beta_4["details"]["commit_def"][
                                            "cid"
                                        ],
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        # officially release the sucessful release candidate to production
        new_version = Version.parse("1.1.0", tag_format=new_version.tag_format)

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEFAULT_BRANCH_NAME},
                },
                {
                    **(merge_beta_into_main_2 := next(merge_beta_into_main_gen)),
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(new_version),
                        "tag_format": new_version.tag_format,
                        "datetime": next(commit_timestamp_gen),
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": new_version,
                                    "dest_files": changelog_file_definitions,
                                    "commit_ids": [
                                        merge_beta_into_main_2["details"]["commit_def"][
                                            "cid"
                                        ],
                                    ],
                                },
                            },
                        ],
                    },
                },
            ]
        )

        return repo_construction_steps

    return _get_repo_from_definition


@pytest.fixture(scope="session")
def build_git_flow_repo_w_4_release_channels(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_git_flow_repo_w_4_release_channels: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_git_flow_repo_w_4_release_channels: str,
) -> BuildSpecificRepoFn:
    def _build_specific_repo_type(
        repo_name: str, commit_type: CommitConvention, dest_dir: Path
    ) -> Sequence[RepoActions]:
        def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
            repo_construction_steps = (
                get_repo_definition_4_git_flow_repo_w_4_release_channels(
                    commit_type=commit_type,
                )
            )
            return build_repo_from_definition(cached_repo_path, repo_construction_steps)

        build_repo_or_copy_cache(
            repo_name=repo_name,
            build_spec_hash=build_spec_hash_4_git_flow_repo_w_4_release_channels,
            build_repo_func=_build_repo,
            dest_dir=dest_dir,
        )

        if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
            raise ValueError("Failed to retrieve repo data from cache")

        return cached_repo_data["build_definition"]

    return _build_specific_repo_type


# --------------------------------------------------------------------------- #
# Test-level fixtures that will cache the built directory & set up test case  #
# --------------------------------------------------------------------------- #


@pytest.fixture
def repo_w_git_flow_w_beta_alpha_rev_prereleases_n_conventional_commits(
    build_git_flow_repo_w_4_release_channels: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = (
        repo_w_git_flow_w_beta_alpha_rev_prereleases_n_conventional_commits.__name__
    )
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_git_flow_repo_w_4_release_channels(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_git_flow_w_beta_alpha_rev_prereleases_n_emoji_commits(
    build_git_flow_repo_w_4_release_channels: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = repo_w_git_flow_w_beta_alpha_rev_prereleases_n_emoji_commits.__name__
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_git_flow_repo_w_4_release_channels(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_git_flow_w_beta_alpha_rev_prereleases_n_scipy_commits(
    build_git_flow_repo_w_4_release_channels: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = repo_w_git_flow_w_beta_alpha_rev_prereleases_n_scipy_commits.__name__
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_git_flow_repo_w_4_release_channels(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }



"""Helper code for interacting with a Bitbucket remote VCS"""

# Note: Bitbucket doesn't support releases. But it allows users to use
# `semantic-release version` without having to specify `--no-vcs-release`.

from __future__ import annotations

import os
from functools import lru_cache
from pathlib import PurePosixPath
from re import compile as regexp
from typing import TYPE_CHECKING

from urllib3.util.url import Url, parse_url

from semantic_release.globals import logger
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Callable


class Bitbucket(RemoteHvcsBase):
    """
    Bitbucket HVCS interface for interacting with BitBucket repositories

    This class supports the following products:

        - BitBucket Cloud
        - BitBucket Data Center Server (on-premises installations)

    This interface does its best to detect which product is configured based
    on the provided domain. If it is the official `bitbucket.org`, the default
    domain, then it is considered as BitBucket Cloud which uses the subdomain
    `api.bitbucket.org/2.0` for api communication.

    If the provided domain is anything else, than it is assumed to be communicating
    with an on-premise or 3rd-party maintained BitBucket instance which matches with
    the BitBucket Data Center Server product. The on-prem server product uses a
    path prefix for handling api requests which is configured to be
    `server.domain/rest/api/1.0` based on the documentation in April 2024.
    """

    OFFICIAL_NAME = "Bitbucket"
    DEFAULT_DOMAIN = "bitbucket.org"
    DEFAULT_API_SUBDOMAIN_PREFIX = "api"
    DEFAULT_API_PATH_CLOUD = "/2.0"
    DEFAULT_API_PATH_ONPREM = "/rest/api/1.0"
    DEFAULT_API_URL_CLOUD = f"https://{DEFAULT_API_SUBDOMAIN_PREFIX}.{DEFAULT_DOMAIN}{DEFAULT_API_PATH_CLOUD}"
    DEFAULT_ENV_TOKEN_NAME = "BITBUCKET_TOKEN"  # noqa: S105

    def __init__(
        self,
        remote_url: str,
        *,
        hvcs_domain: str | None = None,
        hvcs_api_domain: str | None = None,
        token: str | None = None,
        allow_insecure: bool = False,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        super().__init__(remote_url)
        self.token = token
        # NOTE: Uncomment in the future when we actually have functionalty to
        #       use the api, but currently there is none.
        # auth = None if not self.token else TokenAuth(self.token)
        # self.session = build_requests_session(auth=auth)

        domain_url = self._normalize_url(
            hvcs_domain or f"https://{self.DEFAULT_DOMAIN}",
            allow_insecure=allow_insecure,
        )

        # Strip any auth, query or fragment from the domain
        self._hvcs_domain = parse_url(
            Url(
                scheme=domain_url.scheme,
                host=domain_url.host,
                port=domain_url.port,
                path=str(PurePosixPath(domain_url.path or "/")),
            ).url.rstrip("/")
        )

        # Parse api domain if provided otherwise infer from domain
        api_domain_parts = self._normalize_url(
            hvcs_api_domain or self._derive_api_url_from_base_domain(),
            allow_insecure=allow_insecure,
        )

        # As Bitbucket Cloud and Bitbucket Server (on-prem) have different api paths
        # lets check what we have been given and set the api url accordingly
        #   ref: https://developer.atlassian.com/server/bitbucket/how-tos/command-line-rest/
        #   NOTE: BitBucket Server (on premise) uses a path prefix '/rest/api/1.0' for the api
        #         while BitBucket Cloud uses a separate subdomain with '/2.0' path prefix
        is_bitbucket_cloud = bool(
            self.hvcs_domain.url == f"https://{self.DEFAULT_DOMAIN}"
        )

        if (
            is_bitbucket_cloud
            and hvcs_api_domain
            and api_domain_parts.url not in Bitbucket.DEFAULT_API_URL_CLOUD
        ):
            # Api was provied but is not a subset of the expected one, raise an error
            # we check for a subset because the user may not have provided the full api path
            # but the correct domain.  If they didn't, then we are erroring out here.
            raise ValueError(
                f"Invalid api domain {api_domain_parts.url} for BitBucket Cloud. "
                f"Expected {Bitbucket.DEFAULT_API_URL_CLOUD}."
            )

        # Set the api url to the default cloud one if we are on cloud, otherwise
        # use the verified api domain for a on-prem server
        self._api_url = parse_url(
            Bitbucket.DEFAULT_API_URL_CLOUD
            if is_bitbucket_cloud
            else Url(
                # Strip any auth, query or fragment from the domain
                scheme=api_domain_parts.scheme,
                host=api_domain_parts.host,
                port=api_domain_parts.port,
                path=str(
                    PurePosixPath(
                        # pass any custom server prefix path but ensure we don't
                        # double up the api path in the case the user provided it
                        str.replace(
                            api_domain_parts.path or "",
                            self.DEFAULT_API_PATH_ONPREM,
                            "",
                        ).lstrip("/")
                        or "/",
                        # apply the on-prem api path
                        self.DEFAULT_API_PATH_ONPREM.lstrip("/"),
                    )
                ),
            ).url.rstrip("/")
        )

    def _derive_api_url_from_base_domain(self) -> Url:
        return parse_url(
            Url(
                # infer from Domain url and append the api path
                **{
                    **self.hvcs_domain._asdict(),
                    "host": self.hvcs_domain.host,
                    "path": str(
                        PurePosixPath(
                            str.lstrip(self.hvcs_domain.path or "", "/") or "/",
                            self.DEFAULT_API_PATH_ONPREM.lstrip("/"),
                        )
                    ),
                }
            ).url.rstrip("/")
        )

    @lru_cache(maxsize=1)
    def _get_repository_owner_and_name(self) -> tuple[str, str]:
        # ref: https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
        if "BITBUCKET_REPO_FULL_NAME" in os.environ:
            logger.info("Getting repository owner and name from environment variables.")
            owner, name = os.environ["BITBUCKET_REPO_FULL_NAME"].rsplit("/", 1)
            return owner, name

        return super()._get_repository_owner_and_name()

    def remote_url(self, use_token: bool = True) -> str:
        """Get the remote url including the token for authentication if requested"""
        if not use_token:
            return self._remote_url

        if not self.token:
            raise ValueError("Requested to use token but no token set.")

        # If the user is set, assume the token is an user secret. This will work
        # on any repository the user has access to.
        # https://support.atlassian.com/bitbucket-cloud/docs/push-back-to-your-repository
        # If the user variable is not set, assume it is a repository token
        # which will only work on the repository it was created for.
        # https://support.atlassian.com/bitbucket-cloud/docs/using-access-tokens
        user = os.environ.get("BITBUCKET_USER", "x-token-auth")

        return self.create_server_url(
            auth=f"{user}:{self.token}" if user else self.token,
            path=f"/{self.owner}/{self.repo_name}.git",
        )

    def compare_url(self, from_rev: str, to_rev: str) -> str:
        """
        Get the Bitbucket comparison link between two version tags.
        :param from_rev: The older version to compare.
        :param to_rev: The newer version to compare.
        :return: Link to view a comparison between the two versions.
        """
        return self.create_repo_url(
            repo_path=f"/branches/compare/{from_rev}%0D{to_rev}"
        )

    def commit_hash_url(self, commit_hash: str) -> str:
        return self.create_repo_url(repo_path=f"/commits/{commit_hash}")

    def pull_request_url(self, pr_number: str | int) -> str:
        # Strips off any character prefix like '#' that usually exists
        if isinstance(pr_number, str) and (
            match := regexp(r"(\d+)$").search(pr_number)
        ):
            try:
                pr_number = int(match.group(1))
            except ValueError:
                return ""

        if isinstance(pr_number, int):
            return self.create_repo_url(repo_path=f"/pull-requests/{pr_number}")

        return ""

    @staticmethod
    def format_w_official_vcs_name(format_str: str) -> str:
        if "%s" in format_str:
            return format_str % Bitbucket.OFFICIAL_NAME

        if "{}" in format_str:
            return format_str.format(Bitbucket.OFFICIAL_NAME)

        if "{vcs_name}" in format_str:
            return format_str.format(vcs_name=Bitbucket.OFFICIAL_NAME)

        return format_str

    def get_changelog_context_filters(self) -> tuple[Callable[..., Any], ...]:
        return (
            self.create_server_url,
            self.create_repo_url,
            self.commit_hash_url,
            self.compare_url,
            self.pull_request_url,
            self.format_w_official_vcs_name,
        )

    def upload_dists(self, tag: str, dist_glob: str) -> int:
        return super().upload_dists(tag, dist_glob)

    def create_or_update_release(
        self, tag: str, release_notes: str, prerelease: bool = False
    ) -> int | str:
        return super().create_or_update_release(tag, release_notes, prerelease)

    def create_release(
        self,
        tag: str,
        release_notes: str,
        prerelease: bool = False,
        assets: list[str] | None = None,
        noop: bool = False,
    ) -> int | str:
        return super().create_release(tag, release_notes, prerelease, assets, noop)


RemoteHvcsBase.register(Bitbucket)



from enum import Enum
from pathlib import Path

import git

import semantic_release
from semantic_release.cli.commands.main import Cli

PROJ_DIR = Path(__file__).parent.parent.absolute().resolve()


class RepoActionStep(str, Enum):
    CONFIGURE = "CONFIGURE"
    CONFIGURE_MONOREPO = "CONFIGURE_MONOREPO"
    CREATE_MONOREPO = "CREATE_MONOREPO"
    CHANGE_DIRECTORY = "CHANGE_DIRECTORY"
    WRITE_CHANGELOGS = "WRITE_CHANGELOGS"
    GIT_CHECKOUT = "GIT_CHECKOUT"
    GIT_COMMIT = "GIT_COMMIT"
    GIT_MERGE = "GIT_MERGE"
    GIT_SQUASH = "GIT_SQUASH"
    GIT_TAG = "GIT_TAG"
    RELEASE = "RELEASE"
    MAKE_COMMITS = "MAKE_COMMITS"


A_FULL_VERSION_STRING = "1.11.567"
A_PRERELEASE_VERSION_STRING = "2.3.4-dev.23"
A_FULL_VERSION_STRING_WITH_BUILD_METADATA = "4.2.3+build.12345"

EXAMPLE_REPO_OWNER = "example_owner"
EXAMPLE_REPO_NAME = "example_repo"
EXAMPLE_HVCS_DOMAIN = "example.com"

DEFAULT_BRANCH_NAME = "main"
INITIAL_COMMIT_MESSAGE = "Initial commit"

MAIN_PROG_NAME = str(semantic_release.__name__).replace("_", "-")
SUCCESS_EXIT_CODE = 0

CHANGELOG_SUBCMD = Cli.SubCmds.CHANGELOG.name.lower()
GENERATE_CONFIG_SUBCMD = Cli.SubCmds.GENERATE_CONFIG.name.lower()
PUBLISH_SUBCMD = Cli.SubCmds.PUBLISH.name.lower()
VERSION_SUBCMD = Cli.SubCmds.VERSION.name.lower()

NULL_HEX_SHA = git.Object.NULL_HEX_SHA

COMMIT_MESSAGE = "{version}\n\nAutomatically generated by python-semantic-release\n"

SUPPORTED_ISSUE_CLOSURE_PREFIXES = [
    "Close",
    "Closes",
    "Closed",
    "Closing",
    "Fix",
    "Fixes",
    "Fixed",
    "Fixing",
    "Resolve",
    "Resolves",
    "Resolved",
    "Resolving",
    "Implement",
    "Implements",
    "Implemented",
    "Implementing",
]

CONVENTIONAL_COMMITS_CHORE = ("ci: added a commit lint job\n",)
# Different in-scope commits that produce a certain release type
CONVENTIONAL_COMMITS_PATCH = (
    *CONVENTIONAL_COMMITS_CHORE,
    "fix: fixed voltage in the flux capacitor\n",
)
CONVENTIONAL_COMMITS_MINOR = (
    *CONVENTIONAL_COMMITS_PATCH,
    "feat: last minute rush order\n",
)
# Take previous commits and insert a breaking change
CONVENTIONAL_COMMITS_MAJOR = (
    *CONVENTIONAL_COMMITS_MINOR,
    "fix!: big change\n\nBREAKING CHANGE: reworked something for previous feature\n",
)

EMOJI_COMMITS_CHORE = (
    ":broom: updated lint & code style\n",
    ":none: updated ci pipeline\n",
)

EMOJI_COMMITS_PATCH = (
    *EMOJI_COMMITS_CHORE,
    ":bug: fixed voltage in the flux capacitor\n",
)
EMOJI_COMMITS_MINOR = (
    *EMOJI_COMMITS_PATCH,
    ":sparkles::pencil: docs for something special\n",
    # Emoji in description should not be used to evaluate change type
    ":sparkles: last minute rush order\n\nGood thing we're 10x developers :boom:\n",
)
EMOJI_COMMITS_MAJOR = (
    *EMOJI_COMMITS_MINOR,
    ":boom: Move to the blockchain\n",
)

# Note - the scipy commit fixtures for commits that should evaluate to the various scopes
# are in tests/fixtures/scipy

EXAMPLE_PROJECT_NAME = "example"
EXAMPLE_PROJECT_VERSION = "0.0.0"
EXAMPLE_PROJECT_LICENSE = "MIT"

# Uses the internal defaults of semantic-release unless otherwise needed for testing
# modify the pyproject toml as necessary for the test using update_pyproject_toml()
# and derivative fixtures
EXAMPLE_PYPROJECT_TOML_CONTENT = rf"""
[project]
license-expression = "{EXAMPLE_PROJECT_LICENSE}"

[tool.poetry]
name = "{EXAMPLE_PROJECT_NAME}"
version = "{EXAMPLE_PROJECT_VERSION}"
description = "Just an example"
license = "{EXAMPLE_PROJECT_LICENSE}"
authors = ["semantic-release <not-a.real@email.com>"]
readme = "README.md"
classifiers = [
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3 :: Only"
]

[build-system]
requires = ["poetry-core>=1.0.0"]
build-backend = "poetry.core.masonry.api"

[tool.semantic_release]
version_variables = [
    "src/{EXAMPLE_PROJECT_NAME}/_version.py:__version__",
]
version_toml = ["pyproject.toml:tool.poetry.version"]
""".lstrip()

EXAMPLE_SETUP_CFG_CONTENT = rf"""
[metadata]
name = example
version = {EXAMPLE_PROJECT_VERSION}
description = Just an example really
long_description = file: README.md
long_description_content_type = text/markdown
author = semantic-release
author_email = not-a.real@email.com
url = https://github.com/python-semantic-release/python-semantic-release
python_requires = >=3.7


[options]
zip_safe = True
include_package_data = True
packages = find:
install_requires =
    PyYAML==6.0
    pydantic==1.9.0

[options.extras_require]
dev =
    tox
    twine==3.1.1

test =
    pytest
    pytest-cov
    pytest-mock
    pytest-aiohttp

lint =
    flake8
    black>=22.6.0
    isort>=5.10.1

[options.packages.find]
exclude =
    test*

[bdist_wheel]
universal = 1

[coverage:run]
omit = */tests/*

[tools:pytest]
python_files = tests/test_*.py tests/**/test_*.py

[isort]
skip = .tox,venv
default_section = THIRDPARTY
known_first_party = {EXAMPLE_PROJECT_NAME},tests
multi_line_output=3
include_trailing_comma=True
force_grid_wrap=0
use_parentheses=True
line_length=88

[flake8]
max-line-length = 88
""".lstrip()

EXAMPLE_SETUP_PY_CONTENT = rf"""
import re
import sys

from setuptools import find_packages, setup


def _read_long_description():
    try:
        with open("readme.rst") as fd:
            return fd.read()
    except Exception:
        return None


with open("{EXAMPLE_PROJECT_NAME}/_version.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="{EXAMPLE_PROJECT_NAME}",
    version="{EXAMPLE_PROJECT_VERSION}",
    url="http://github.com/python-semantic-release/python-semantic-release",
    author="semantic-release",
    author_email="not-a.real@email.com",
    description="Just an example",
    long_description=_read_long_description(),
    packages=find_packages(exclude=("tests",)),
    license="MIT",
    install_requires=[
        "click>=7,<9",
        "click_log>=0.3,<1",
        "gitpython>=3.0.8,<4",
        "invoke>=1.4.1,<2",
        "semver>=2.10,<3",
        "twine>=3,<4",
        "requests>=2.25,<3",
        "wheel",
        "python-gitlab>=2,<4",
        # tomlkit used to be pinned to 0.7.0
        # See https://github.com/python-semantic-release/python-semantic-release/issues/336
        # and https://github.com/python-semantic-release/python-semantic-release/pull/337
        # and https://github.com/python-semantic-release/python-semantic-release/issues/491
        "tomlkit~=0.10",
        "dotty-dict>=1.3.0,<2",
        "dataclasses==0.8; python_version < '3.7.0'",
        "packaging",
    ],
    extras_require={{
        "test": [
            "coverage>=5,<6",
            "pytest>=5,<6",
            "pytest-xdist>=1,<2",
            "pytest-mock>=2,<3",
            "pytest-lazy-fixture~=0.6.3",
            "responses==0.13.3",
            "mock==1.3.0",
        ],
        "docs": ["Sphinx==1.3.6", "Jinja2==3.0.3"],
        "dev": ["tox", "isort", "black"],
        "mypy": ["mypy", "types-requests"],
    }},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
""".lstrip()

EXAMPLE_CHANGELOG_MD_CONTENT = r"""
# CHANGELOG
<!-- This is an example changelog -->

## v1.0.0

* Various bugfixes, security enhancements
* Extra cookies to enhance your experience
* ~Removed~ simplified cookie opt-out handling logic

""".lstrip()

EXAMPLE_CHANGELOG_RST_CONTENT = r"""
.. _changelog:
=========
CHANGELOG
=========

..
  example project base changelog

.. _changelog-v1.0.0:

v1.0.0 (1970-01-01)
===================

* Various bugfixes, security enhancements
* Extra cookies to enhance your experience
* ~Removed~ simplified cookie opt-out handling logic
""".lstrip()

EXAMPLE_RELEASE_NOTES_TEMPLATE = r"""
## What's Changed
{%    for type_, commits in release["elements"] | dictsort
%}{{    "### %s" | format(type_ | title)
}}{%    if type_ != "unknown"
%}{%      for commit in commits
%}{{        "* %s" | format(commit.descriptions[0] | trim)
}}{%      endfor
%}{%    endif
%}{%  endfor
%}
""".lstrip()  # noqa: E501

RELEASE_NOTES = "# Release Notes"
DEFAULT_MERGE_STRATEGY_OPTION = "theirs"



from __future__ import annotations

import logging
from enum import IntEnum, unique


@unique
class LevelBump(IntEnum):
    """
    IntEnum representing valid types of bumps for a version.
    We use an IntEnum to enable ordering of levels.
    """

    NO_RELEASE = 0
    PRERELEASE_REVISION = 1
    PATCH = 2
    MINOR = 3
    MAJOR = 4

    def __str__(self) -> str:
        """
        Return the level name rather than 'LevelBump.<level>'
        E.g.
        >>> str(LevelBump.NO_RELEASE)
        'no_release'
        >>> str(LevelBump.MAJOR)
        'major'
        """
        return self.name.lower()

    @classmethod
    def from_string(cls, val: str) -> LevelBump:
        """
        Get the level from string representation. For backwards-compatibility,
        dashes are replaced with underscores so that:
        >>> LevelBump.from_string("no-release") == LevelBump.NO_RELEASE
        Equally,
        >>> LevelBump.from_string("minor") == LevelBump.MINOR
        """
        return cls[val.upper().replace("-", "_")]


class SemanticReleaseLogLevels(IntEnum):
    """IntEnum representing the log levels used by semantic-release."""

    FATAL = logging.FATAL
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    SILLY = 5

    def __str__(self) -> str:
        """
        Return the level name rather than 'SemanticReleaseLogLevels.<level>'
        E.g.
        >>> str(SemanticReleaseLogLevels.DEBUG)
        'DEBUG'
        >>> str(SemanticReleaseLogLevels.CRITICAL)
        'CRITICAL'
        """
        return self.name.upper()


logging.addLevelName(
    SemanticReleaseLogLevels.SILLY,
    str(SemanticReleaseLogLevels.SILLY),
)



from __future__ import annotations

import importlib.util
import os
import re
import string
import sys
from functools import lru_cache, reduce, wraps
from pathlib import Path, PurePosixPath
from re import IGNORECASE, compile as regexp
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Sequence, TypeVar
from urllib.parse import urlsplit

from semantic_release.globals import logger

if TYPE_CHECKING:  # pragma: no cover
    from logging import Logger
    from re import Pattern
    from typing import Iterable


number_pattern = regexp(r"(?P<prefix>\S*?)(?P<number>\d[\d,]*)\b")
hex_number_pattern = regexp(
    r"(?P<prefix>\S*?)(?:0x)?(?P<number>[0-9a-f]+)\b", IGNORECASE
)


def get_number_from_str(
    string: str, default: int = -1, interpret_hex: bool = False
) -> int:
    if interpret_hex and (match := hex_number_pattern.search(string)):
        return abs(int(match.group("number"), 16))

    if match := number_pattern.search(string):
        return int(match.group("number"))

    return default


def sort_numerically(
    iterable: Iterable[str], reverse: bool = False, allow_hex: bool = False
) -> list[str]:
    # Alphabetically sort prefixes first, then sort by number
    alphabetized_list = sorted(iterable)

    # Extract prefixes in order to group items by prefix
    unmatched_items = []
    prefixes: dict[str, list[str]] = {}
    for item in alphabetized_list:
        if not (
            pattern_match := (
                (hex_number_pattern.search(item) if allow_hex else None)
                or number_pattern.search(item)
            )
        ):
            unmatched_items.append(item)
            continue

        prefix = prefix if (prefix := pattern_match.group("prefix")) else ""

        if prefix not in prefixes:
            prefixes[prefix] = []

        prefixes[prefix].append(item)

    # Sort prefixes and items by number mixing in unmatched items as alphabetized with other prefixes
    return reduce(
        lambda acc, next_item: acc + next_item,
        [
            (
                sorted(
                    prefixes[prefix],
                    key=lambda x: get_number_from_str(
                        x, default=-1, interpret_hex=allow_hex
                    ),
                    reverse=reverse,
                )
                if prefix in prefixes
                else [prefix]
            )
            for prefix in sorted([*prefixes.keys(), *unmatched_items])
        ],
        [],
    )


def text_reducer(text: str, filter_pair: tuple[Pattern[str], str]) -> str:
    """Reduce function to apply mulitple filters to a string"""
    if not text:  # abort if the paragraph is empty
        return text

    filter_pattern, replacement = filter_pair
    return filter_pattern.sub(replacement, text)


def validate_types_in_sequence(
    sequence: Sequence, types: type | tuple[type, ...]
) -> bool:
    """Validate that all elements in a sequence are of a specific type"""
    return all(isinstance(item, types) for item in sequence)


def format_arg(value: Any) -> str:
    """Helper to format an argument an argument for logging"""
    if type(value) == str:
        return f"'{value.strip()}'"
    return str(value)


def check_tag_format(tag_format: str) -> None:
    if "version" not in (f[1] for f in string.Formatter().parse(tag_format)):
        raise ValueError(
            f"Invalid tag_format {tag_format!r}, must use 'version' as a format key"
        )


_R = TypeVar("_R")
_FuncType = Callable[..., _R]


def logged_function(logger: Logger) -> Callable[[_FuncType[_R]], _FuncType[_R]]:
    """
    Decorator which adds debug logging of a function's input arguments and return
    value.

    The input arguments are logged before the function is called, and the
    return value is logged once it has completed.

    :param logger: Logger to send output to.
    """

    def _logged_function(func: _FuncType[_R]) -> _FuncType[_R]:
        @wraps(func)
        def _wrapper(*args: Any, **kwargs: Any) -> _R:
            logger.debug(
                "%s(%s, %s)",
                func.__name__,
                ", ".join([format_arg(x) for x in args]),
                ", ".join([f"{k}={format_arg(v)}" for k, v in kwargs.items()]),
            )

            # Call function
            result = func(*args, **kwargs)

            # Log result
            logger.debug("%s -> %s", func.__qualname__, str(result))
            return result

        return _wrapper

    return _logged_function


@logged_function(logger)
def dynamic_import(import_path: str) -> Any:
    """
    Dynamically import an object from a conventionally formatted "module:attribute"
    string
    """
    if ":" not in import_path:
        raise ValueError(
            f"Invalid import path {import_path!r}, must use 'module:Class' format"
        )

    # Split the import path into module and attribute
    module_name, attr = import_path.split(":", maxsplit=1)

    # Check if the module is a file path, if it can be resolved and exists on disk then import as a file
    module_filepath = Path(module_name).resolve()
    if module_filepath.exists():
        module_path = (
            module_filepath.stem
            if Path(module_name).is_absolute()
            else str(Path(module_name).with_suffix("")).replace(os.sep, ".").lstrip(".")
        )

        if module_path not in sys.modules:
            logger.debug("Loading '%s' from file '%s'", module_path, module_filepath)
            spec = importlib.util.spec_from_file_location(
                module_path, str(module_filepath)
            )
            if spec is None:
                raise ImportError(f"Could not import {module_filepath}")

            module = importlib.util.module_from_spec(spec)  # type: ignore[arg-type]
            sys.modules.update({spec.name: module})
            spec.loader.exec_module(module)  # type: ignore[union-attr]

        return getattr(sys.modules[module_path], attr)

    # Otherwise, import as a module
    try:
        logger.debug("Importing module '%s'", module_name)
        module = importlib.import_module(module_name)
        logger.debug("Loading '%s' from module '%s'", attr, module_name)
        return getattr(module, attr)
    except TypeError as err:
        raise ImportError(
            str.join(
                "\n",
                [
                    str(err.args[0]),
                    "Verify the import format matches 'module:attribute' or 'path/to/module:attribute'",
                ],
            )
        ) from err


class ParsedGitUrl(NamedTuple):
    """Container for the elements parsed from a git URL"""

    scheme: str
    netloc: str
    namespace: str
    repo_name: str


@lru_cache(maxsize=512)
def parse_git_url(url: str) -> ParsedGitUrl:
    """
    Attempt to parse a string as a git url http[s]://, git://, file://, or ssh format, into a
    ParsedGitUrl.

    supported examples:
        http://git.mycompany.com/username/myproject.git
        https://github.com/username/myproject.git
        https://gitlab.com/group/subgroup/myproject.git
        https://git.mycompany.com:4443/username/myproject.git
        git://host.xz/path/to/repo.git/
        git://host.xz:9418/path/to/repo.git/
        git@github.com:username/myproject.git                  <-- assumes ssh://
        ssh://git@github.com:3759/myproject.git                <-- non-standard, but assume user 3759
        ssh://git@github.com:username/myproject.git
        ssh://git@bitbucket.org:7999/username/myproject.git
        git+ssh://git@github.com:username/myproject.git
        /Users/username/dev/remote/myproject.git               <-- Posix File paths
        file:///Users/username/dev/remote/myproject.git
        C:/Users/username/dev/remote/myproject.git             <-- Windows File paths
        file:///C:/Users/username/dev/remote/myproject.git

    REFERENCE: https://stackoverflow.com/questions/31801271/what-are-the-supported-git-url-formats

    Raises ValueError if the url can't be parsed.
    """
    logger.debug("Parsing git url %r", url)

    # Normalizers are a list of tuples of (pattern, replacement)
    normalizers = [
        # normalize implicit ssh urls to explicit ssh://
        (r"^([\w._-]+@)", r"ssh://\1"),
        # normalize git+ssh:// urls to ssh://
        (r"^git\+ssh://", "ssh://"),
        # normalize an scp like syntax to URL compatible syntax
        # excluding port definitions (:#####) & including numeric usernames
        (r"(ssh://(?:[\w._-]+@)?[\w.-]+):(?!\d{1,5}/\w+/)(.*)$", r"\1/\2"),
        # normalize implicit file (windows || posix) urls to explicit file:// urls
        (r"^([C-Z]:/)|^/(\w)", r"file:///\1\2"),
    ]

    for pattern, replacement in normalizers:
        url = re.compile(pattern).sub(replacement, url)

    # run the url through urlsplit to separate out the parts
    urllib_split = urlsplit(url)

    # Fail if url scheme not found
    if not urllib_split.scheme:
        raise ValueError(f"Cannot parse {url!r}")

    # We have been able to parse the url with urlsplit,
    # so it's a (file|git|ssh|https?)://... structure
    # but we aren't validating the protocol scheme as its not our business

    # use PosixPath to normalize the path & then separate out the namespace & repo_name
    namespace, _, name = (
        str(PurePosixPath(urllib_split.path)).lstrip("/").rpartition("/")
    )

    # strip out the .git at the end of the repo_name if present
    name = name[:-4] if name.endswith(".git") else name

    # check that we have all the required parts of the url
    required_parts = [
        urllib_split.scheme,
        # Allow empty net location for file:// urls
        True if urllib_split.scheme == "file" else urllib_split.netloc,
        namespace,
        name,
    ]

    if not all(required_parts):
        raise ValueError(f"Bad url: {url!r}")

    return ParsedGitUrl(
        scheme=urllib_split.scheme,
        netloc=urllib_split.netloc,
        namespace=namespace,
        repo_name=name,
    )



import semantic_release.version.declaration as declaration
from semantic_release.version.algorithm import (
    next_version,
    tags_and_versions,
)
from semantic_release.version.translator import VersionTranslator
from semantic_release.version.version import Version



from __future__ import annotations

import os
from fnmatch import fnmatch
from logging import getLogger
from pathlib import Path, PurePath, PurePosixPath, PureWindowsPath
from re import DOTALL, compile as regexp, error as RegexError  # noqa: N812
from typing import TYPE_CHECKING

from semantic_release.commit_parser._base import CommitParser
from semantic_release.commit_parser.conventional.options import (
    ConventionalCommitParserOptions,
)
from semantic_release.commit_parser.conventional.options_monorepo import (
    ConventionalCommitMonorepoParserOptions,
)
from semantic_release.commit_parser.conventional.parser import ConventionalCommitParser
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParsedMessageResult,
    ParseError,
    ParseResult,
)
from semantic_release.commit_parser.util import force_str
from semantic_release.errors import InvalidParserOptions

if TYPE_CHECKING:  # pragma: no cover
    from git.objects.commit import Commit


class ConventionalCommitMonorepoParser(
    CommitParser[ParseResult, ConventionalCommitMonorepoParserOptions]
):
    # TODO: Remove for v11 compatibility, get_default_options() will be called instead
    parser_options = ConventionalCommitMonorepoParserOptions

    def __init__(
        self, options: ConventionalCommitMonorepoParserOptions | None = None
    ) -> None:
        super().__init__(options)

        try:
            commit_scope_pattern = regexp(
                r"\(" + self.options.scope_prefix + r"(?P<scope>[^\n]+)?\)",
            )
        except RegexError as err:
            raise InvalidParserOptions(
                str.join(
                    "\n",
                    [
                        f"Invalid options for {self.__class__.__name__}",
                        "Unable to create regular expression from configured scope_prefix.",
                        "Please check the configured scope_prefix and remove or escape any regular expression characters.",
                    ],
                )
            ) from err

        try:
            commit_type_pattern = regexp(
                r"(?P<type>%s)" % str.join("|", self.options.allowed_tags)
            )
        except RegexError as err:
            raise InvalidParserOptions(
                str.join(
                    "\n",
                    [
                        f"Invalid options for {self.__class__.__name__}",
                        "Unable to create regular expression from configured commit-types.",
                        "Please check the configured commit-types and remove or escape any regular expression characters.",
                    ],
                )
            ) from err

        # This regular expression includes scope prefix into the pattern and forces a scope to be present
        # PSR will match the full scope but we don't include it in the scope match,
        # which implicitly strips it from being included in the returned scope.
        self._strict_scope_pattern = regexp(
            str.join(
                "",
                [
                    r"^" + commit_type_pattern.pattern,
                    commit_scope_pattern.pattern,
                    r"(?P<break>!)?:\s+",
                    r"(?P<subject>[^\n]+)",
                    r"(?:\n\n(?P<text>.+))?",  # commit body
                ],
            ),
            flags=DOTALL,
        )

        self._optional_scope_pattern = regexp(
            str.join(
                "",
                [
                    r"^" + commit_type_pattern.pattern,
                    r"(?:\((?P<scope>[^\n]+)\))?",
                    r"(?P<break>!)?:\s+",
                    r"(?P<subject>[^\n]+)",
                    r"(?:\n\n(?P<text>.+))?",  # commit body
                ],
            ),
            flags=DOTALL,
        )

        file_select_filters, file_ignore_filters = self._process_path_filter_options(
            self.options.path_filters
        )
        self._file_selection_filters: list[str] = file_select_filters
        self._file_ignore_filters: list[str] = file_ignore_filters

        self._logger = getLogger(
            str.join(".", [self.__module__, self.__class__.__name__])
        )

        self._base_parser = ConventionalCommitParser(
            options=ConventionalCommitParserOptions(
                **{
                    k: getattr(self.options, k)
                    for k in ConventionalCommitParserOptions().__dataclass_fields__
                }
            )
        )

    def get_default_options(self) -> ConventionalCommitMonorepoParserOptions:
        return ConventionalCommitMonorepoParserOptions()

    @staticmethod
    def _process_path_filter_options(  # noqa: C901
        path_filters: tuple[str, ...],
    ) -> tuple[list[str], list[str]]:
        file_ignore_filters: list[str] = []
        file_selection_filters: list[str] = []
        unique_selection_filters: set[str] = set()
        unique_ignore_filters: set[str] = set()

        for str_path in path_filters:
            str_filter = str_path[1:] if str_path.startswith("!") else str_path
            filter_list = (
                file_ignore_filters
                if str_path.startswith("!")
                else file_selection_filters
            )
            unique_cache = (
                unique_ignore_filters
                if str_path.startswith("!")
                else unique_selection_filters
            )

            # Since fnmatch is not too flexible, we will expand the path filters to include the name and any subdirectories
            # as this is how gitignore is interpreted. Possible scenarios:
            # | Input      | Path Normalization | Filter List               |
            # | ---------- | ------------------ | ------------------------- |
            # | /          | /                  | /**                       | done
            # | /./        | /                  | /**                       | done
            # | /**        | /**                | /**                       | done
            # | /./**      | /**                | /**                       | done
            # | /*         | /*                 | /*                        | done
            # | .          | .                  | ./**                      | done
            # | ./         | .                  | ./**                      | done
            # | ././       | .                  | ./**                      | done
            # | ./**       | ./**               | ./**                      | done
            # | ./*        | ./*                | ./*                       | done
            # | ..         | ..                 | ../**                     | done
            # | ../        | ..                 | ../**                     | done
            # | ../**      | ../**              | ../**                     | done
            # | ../*       | ../*               | ../*                      | done
            # | ../..      | ../..              | ../../**                  | done
            # | ../../     | ../../             | ../../**                  | done
            # | ../../docs | ../../docs         | ../../docs, ../../docs/** | done
            # | src        | src                | src, src/**               | done
            # | src/       | src                | src/**                    | done
            # | src/*      | src/*              | src/*                     | done
            # | src/**     | src/**             | src/**                    | done
            # | /src       | /src               | /src, /src/**             | done
            # | /src/      | /src               | /src/**                   | done
            # | /src/**    | /src/**            | /src/**                   | done
            # | /src/*     | /src/*             | /src/*                    | done
            # | ../d/f.txt | ../d/f.txt         | ../d/f.txt, ../d/f.txt/** | done
            # This expansion will occur regardless of the negation prefix

            os_path: PurePath | PurePosixPath | PureWindowsPath = PurePath(str_filter)

            if r"\\" in str_filter:
                # Windows paths were given so we convert them to posix paths
                os_path = PureWindowsPath(str_filter)
                os_path = (
                    PureWindowsPath(
                        os_path.root, *os_path.parts[1:]
                    )  # drop any drive letter
                    if os_path.is_absolute()
                    else os_path
                )
                os_path = PurePosixPath(os_path.as_posix())

            path_normalized = str(os_path)
            if path_normalized == str(
                Path(".").absolute().root
            ) or path_normalized == str(Path("/**")):
                path_normalized = "/**"

            elif path_normalized == str(Path("/*")):
                pass

            elif path_normalized == str(Path(".")) or path_normalized == str(
                Path("./**")
            ):
                path_normalized = "./**"

            elif path_normalized == str(Path("./*")):
                path_normalized = "./*"

            elif path_normalized == str(Path("..")) or path_normalized == str(
                Path("../**")
            ):
                path_normalized = "../**"

            elif path_normalized == str(Path("../*")):
                path_normalized = "../*"

            elif path_normalized.endswith(("..", "../**")):
                path_normalized = f"{path_normalized.rstrip('*')}/**"

            elif str_filter.endswith(os.sep):
                # If the path ends with a separator, it is a directory, so we add the directory and all subdirectories
                path_normalized = f"{path_normalized}/**"

            elif not path_normalized.endswith("*"):
                all_subdirs = f"{path_normalized}/**"
                if all_subdirs not in unique_cache:
                    unique_cache.add(all_subdirs)
                    filter_list.append(all_subdirs)
                # And fall through to add the path as is

            # END IF

            # Add the normalized path to the filter list if it is not already present
            if path_normalized not in unique_cache:
                unique_cache.add(path_normalized)
                filter_list.append(path_normalized)

        return file_selection_filters, file_ignore_filters

    def logged_parse_error(self, commit: Commit, error: str) -> ParseError:
        self._logger.debug(error)
        return ParseError(commit, error=error)

    def parse(self, commit: Commit) -> ParseResult | list[ParseResult]:
        if self.options.ignore_merge_commits and self._base_parser.is_merge_commit(
            commit
        ):
            return self._base_parser.log_parse_error(
                commit, "Ignoring merge commit: %s" % commit.hexsha[:8]
            )

        separate_commits: list[Commit] = (
            self._base_parser.unsquash_commit(commit)
            if self.options.parse_squash_commits
            else [commit]
        )

        # Parse each commit individually if there were more than one
        parsed_commits: list[ParseResult] = list(
            map(self.parse_commit, separate_commits)
        )

        def add_linked_merge_request(
            parsed_result: ParseResult, mr_number: str
        ) -> ParseResult:
            return (
                parsed_result
                if not isinstance(parsed_result, ParsedCommit)
                else ParsedCommit(
                    **{
                        **parsed_result._asdict(),
                        "linked_merge_request": mr_number,
                    }
                )
            )

        # TODO: improve this for other VCS systems other than GitHub & BitBucket
        # Github works as the first commit in a squash merge commit has the PR number
        # appended to the first line of the commit message
        lead_commit = next(iter(parsed_commits))

        if isinstance(lead_commit, ParsedCommit) and lead_commit.linked_merge_request:
            # If the first commit has linked merge requests, assume all commits
            # are part of the same PR and add the linked merge requests to all
            # parsed commits
            parsed_commits = [
                lead_commit,
                *map(
                    lambda parsed_result, mr=lead_commit.linked_merge_request: (  # type: ignore[misc]
                        add_linked_merge_request(parsed_result, mr)
                    ),
                    parsed_commits[1:],
                ),
            ]

        elif isinstance(lead_commit, ParseError) and (
            mr_match := self._base_parser.mr_selector.search(
                force_str(lead_commit.message)
            )
        ):
            # Handle BitBucket Squash Merge Commits (see #1085), which have non angular commit
            # format but include the PR number in the commit subject that we want to extract
            linked_merge_request = mr_match.group("mr_number")

            # apply the linked MR to all commits
            parsed_commits = [
                add_linked_merge_request(parsed_result, linked_merge_request)
                for parsed_result in parsed_commits
            ]

        return parsed_commits

    def parse_message(
        self, message: str, strict_scope: bool = False
    ) -> ParsedMessageResult | None:
        if (
            not (parsed_match := self._strict_scope_pattern.match(message))
            and strict_scope
        ):
            return None

        if not parsed_match and not (
            parsed_match := self._optional_scope_pattern.match(message)
        ):
            return None

        return self._base_parser.create_parsed_message_result(parsed_match)

    def parse_commit(self, commit: Commit) -> ParseResult:
        """Attempt to parse the commit message with a regular expression into a ParseResult."""
        # Multiple scenarios to consider when parsing a commit message [Truth table]:
        # =======================================================================================================
        # |    ||                         INPUTS                         ||                                     |
        # |  # ||------------------------+----------------+--------------||                Result               |
        # |    || Example Commit Message | Relevant Files | Scope Prefix ||                                     |
        # |----||------------------------+----------------+--------------||-------------------------------------|
        # |  1 || type(prefix-cli): msg  |            yes |    "prefix-" ||                        ParsedCommit |
        # |  2 || type(prefix-cli): msg  |            yes |           "" ||                        ParsedCommit |
        # |  3 || type(prefix-cli): msg  |             no |    "prefix-" ||                        ParsedCommit |
        # |  4 || type(prefix-cli): msg  |             no |           "" ||                ParseError[No files] |
        # |  5 || type(scope-cli): msg   |            yes |    "prefix-" ||                        ParsedCommit |
        # |  6 || type(scope-cli): msg   |            yes |           "" ||                        ParsedCommit |
        # |  7 || type(scope-cli): msg   |             no |    "prefix-" ||  ParseError[No files & wrong scope] |
        # |  8 || type(scope-cli): msg   |             no |           "" ||                ParseError[No files] |
        # |  9 || type(cli): msg         |            yes |    "prefix-" ||                        ParsedCommit |
        # | 10 || type(cli): msg         |            yes |           "" ||                        ParsedCommit |
        # | 11 || type(cli): msg         |             no |    "prefix-" ||  ParseError[No files & wrong scope] |
        # | 12 || type(cli): msg         |             no |           "" ||                ParseError[No files] |
        # | 13 || type: msg              |            yes |    "prefix-" ||                        ParsedCommit |
        # | 14 || type: msg              |            yes |           "" ||                        ParsedCommit |
        # | 15 || type: msg              |             no |    "prefix-" ||  ParseError[No files & wrong scope] |
        # | 16 || type: msg              |             no |           "" ||                ParseError[No files] |
        # | 17 || non-conventional msg   |            yes |    "prefix-" ||          ParseError[Invalid Syntax] |
        # | 18 || non-conventional msg   |            yes |           "" ||          ParseError[Invalid Syntax] |
        # | 19 || non-conventional msg   |             no |    "prefix-" ||          ParseError[Invalid Syntax] |
        # | 20 || non-conventional msg   |             no |           "" ||          ParseError[Invalid Syntax] |
        # =======================================================================================================

        # Initial Logic Flow:
        # [1] When there are no relevant files and a scope prefix is defined, we enforce a strict scope
        # [2] When there are no relevant files and no scope prefix is defined, we parse scoped or unscoped commits
        # [3] When there are relevant files, we parse scoped or unscoped commits regardless of any defined prefix
        has_relevant_changed_files = self._has_relevant_changed_files(commit)
        strict_scope = bool(
            not has_relevant_changed_files and self.options.scope_prefix
        )
        pmsg_result = self.parse_message(
            message=force_str(commit.message),
            strict_scope=strict_scope,
        )

        if pmsg_result and (has_relevant_changed_files or strict_scope):
            self._logger.debug(
                "commit %s introduces a %s level_bump",
                commit.hexsha[:8],
                pmsg_result.bump,
            )

            return ParsedCommit.from_parsed_message_result(commit, pmsg_result)

        if pmsg_result and not has_relevant_changed_files:
            return self.logged_parse_error(
                commit,
                f"Commit {commit.hexsha[:7]} has no changed files matching the path filter(s)",
            )

        if strict_scope and self.parse_message(str(commit.message), strict_scope=False):
            return self.logged_parse_error(
                commit,
                str.join(
                    " and ",
                    [
                        f"Commit {commit.hexsha[:7]} has no changed files matching the path filter(s)",
                        f"the scope does not match scope prefix '{self.options.scope_prefix}'",
                    ],
                ),
            )

        return self.logged_parse_error(
            commit,
            f"Format Mismatch! Unable to parse commit message: {commit.message!r}",
        )

    def unsquash_commit_message(self, message: str) -> list[str]:
        return self._base_parser.unsquash_commit_message(message)

    def _has_relevant_changed_files(self, commit: Commit) -> bool:
        # Extract git root from commit
        git_root = (
            Path(commit.repo.working_tree_dir or commit.repo.working_dir)
            .absolute()
            .resolve()
        )

        cwd = Path.cwd().absolute().resolve()

        rel_cwd = cwd.relative_to(git_root) if git_root in cwd.parents else Path(".")

        sandboxed_selection_filters: list[str] = [
            str(file_filter)
            for file_filter in (
                (
                    git_root / select_filter.rstrip("/")
                    if Path(select_filter).is_absolute()
                    else git_root / rel_cwd / select_filter
                )
                for select_filter in self._file_selection_filters
            )
            if git_root in file_filter.parents
        ]

        sandboxed_ignore_filters: list[str] = [
            str(file_filter)
            for file_filter in (
                (
                    git_root / ignore_filter.rstrip("/")
                    if Path(ignore_filter).is_absolute()
                    else git_root / rel_cwd / ignore_filter
                )
                for ignore_filter in self._file_ignore_filters
            )
            if git_root in file_filter.parents
        ]

        # Check if the changed files of the commit that match the path filters
        for full_path in iter(
            str(git_root / rel_git_path) for rel_git_path in commit.stats.files
        ):
            # Check if the filepath matches any of the file selection filters
            if not any(
                fnmatch(full_path, select_filter)
                for select_filter in sandboxed_selection_filters
            ):
                continue

            # Pass filter matches, so now evaluate if it is supposed to be ignored
            if not any(
                fnmatch(full_path, ignore_filter)
                for ignore_filter in sandboxed_ignore_filters
            ):
                # No ignore filter matched, so it must be a relevant file
                return True

        return False



"""Helper code for interacting with a Gitea remote VCS"""

from __future__ import annotations

import glob
import os
from pathlib import PurePosixPath
from re import compile as regexp
from typing import TYPE_CHECKING

from requests import HTTPError, JSONDecodeError
from urllib3.util.url import Url, parse_url

from semantic_release.cli.util import noop_report
from semantic_release.errors import (
    AssetUploadError,
    IncompleteReleaseError,
    UnexpectedResponse,
)
from semantic_release.globals import logger
from semantic_release.helpers import logged_function
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase
from semantic_release.hvcs.token_auth import TokenAuth
from semantic_release.hvcs.util import build_requests_session, suppress_not_found

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Callable


class Gitea(RemoteHvcsBase):
    """Gitea helper class"""

    OFFICIAL_NAME = "Gitea"
    DEFAULT_DOMAIN = "gitea.com"
    DEFAULT_API_PATH = "/api/v1"
    DEFAULT_ENV_TOKEN_NAME = "GITEA_TOKEN"  # noqa: S105

    def __init__(
        self,
        remote_url: str,
        *,
        hvcs_domain: str | None = None,
        token: str | None = None,
        allow_insecure: bool = False,
        **_kwargs: Any,
    ) -> None:
        super().__init__(remote_url)
        self.token = token
        auth = None if not self.token else TokenAuth(self.token)
        self.session = build_requests_session(auth=auth)

        domain_url = self._normalize_url(
            hvcs_domain
            or os.getenv("GITEA_SERVER_URL", "")
            or f"https://{self.DEFAULT_DOMAIN}",
            allow_insecure=allow_insecure,
        )

        # Strip any auth, query or fragment from the domain
        self._hvcs_domain = parse_url(
            Url(
                scheme=domain_url.scheme,
                host=domain_url.host,
                port=domain_url.port,
                path=str(PurePosixPath(domain_url.path or "/")),
            ).url.rstrip("/")
        )

        self._api_url = self._normalize_url(
            os.getenv("GITEA_API_URL", "").rstrip("/")
            or Url(
                # infer from Domain url and append the default api path
                **{
                    **self.hvcs_domain._asdict(),
                    "path": f"{self.hvcs_domain.path or ''}{self.DEFAULT_API_PATH}",
                }
            ).url,
            allow_insecure=allow_insecure,
        )

    @logged_function(logger)
    def create_release(
        self,
        tag: str,
        release_notes: str,
        prerelease: bool = False,
        assets: list[str] | None = None,
        noop: bool = False,
    ) -> int:
        """
        Create a new release

        Ref: https://gitea.com/api/swagger#/repository/repoCreateRelease

        :param tag: Tag to create release for
        :param release_notes: The release notes for this version
        :param prerelease: Whether or not this release should be specified as a
            prerelease

        :return: Whether the request succeeded
        """
        if noop:
            noop_report(
                str.join(
                    " ",
                    [
                        f"would have created a release for tag {tag}",
                        "with the following notes:\n",
                        release_notes,
                    ],
                )
            )
            if assets:
                noop_report(
                    str.join(
                        "\n",
                        [
                            "would have uploaded the following assets to the release:",
                            *assets,
                        ],
                    )
                )
            return -1

        logger.info("Creating release for tag %s", tag)
        releases_endpoint = self.create_api_url(
            endpoint=f"/repos/{self.owner}/{self.repo_name}/releases",
        )
        response = self.session.post(
            releases_endpoint,
            json={
                "tag_name": tag,
                "name": tag,
                "body": release_notes,
                "draft": False,
                "prerelease": prerelease,
            },
        )

        # Raise an error if the request was not successful
        response.raise_for_status()

        try:
            release_id: int = response.json()["id"]
            logger.info("Successfully created release with ID: %s", release_id)
        except JSONDecodeError as err:
            raise UnexpectedResponse("Unreadable json response") from err
        except KeyError as err:
            raise UnexpectedResponse("JSON response is missing an id") from err

        errors = []
        for asset in assets or []:
            logger.info("Uploading asset %s", asset)
            try:
                self.upload_release_asset(release_id, asset)
            except HTTPError as err:
                errors.append(
                    AssetUploadError(f"Failed asset upload for {asset}").with_traceback(
                        err.__traceback__
                    )
                )

        if len(errors) < 1:
            return release_id

        for error in errors:
            logger.exception(error)

        raise IncompleteReleaseError(
            f"Failed to upload asset{'s' if len(errors) > 1 else ''} to release!"
        )

    @logged_function(logger)
    @suppress_not_found
    def get_release_id_by_tag(self, tag: str) -> int | None:
        """
        Get a release by its tag name
        https://gitea.com/api/swagger#/repository/repoGetReleaseByTag
        :param tag: Tag to get release for

        :return: ID of found release
        """
        tag_endpoint = self.create_api_url(
            endpoint=f"/repos/{self.owner}/{self.repo_name}/releases/tags/{tag}",
        )
        response = self.session.get(tag_endpoint)

        # Raise an error if the request was not successful
        response.raise_for_status()

        try:
            data = response.json()
            return data["id"]
        except JSONDecodeError as err:
            raise UnexpectedResponse("Unreadable json response") from err
        except KeyError as err:
            raise UnexpectedResponse("JSON response is missing an id") from err

    @logged_function(logger)
    def edit_release_notes(self, release_id: int, release_notes: str) -> int:
        """
        Edit a release with updated change notes
        https://gitea.com/api/swagger#/repository/repoEditRelease
        :param id: ID of release to update
        :param release_notes: The release notes for this version

        :return: The ID of the release that was edited
        """
        logger.info("Updating release %s", release_id)
        release_endpoint = self.create_api_url(
            endpoint=f"/repos/{self.owner}/{self.repo_name}/releases/{release_id}",
        )

        response = self.session.patch(
            release_endpoint,
            json={"body": release_notes},
        )

        # Raise an error if the request was not successful
        response.raise_for_status()

        return release_id

    @logged_function(logger)
    def create_or_update_release(
        self, tag: str, release_notes: str, prerelease: bool = False
    ) -> int:
        """
        Post release changelog
        :param version: The version number
        :param changelog: The release notes for this version

        :return: The status of the request
        """
        logger.info("Creating release for %s", tag)
        try:
            return self.create_release(tag, release_notes, prerelease)
        except HTTPError as err:
            logger.debug("error creating release: %s", err)
            logger.debug("looking for an existing release to update")

        release_id = self.get_release_id_by_tag(tag)
        if release_id is None:
            raise ValueError(
                f"release id for tag {tag} not found, and could not be created"
            )

        # If this errors we let it die
        logger.debug("Found existing release %s, updating", release_id)
        return self.edit_release_notes(release_id, release_notes)

    @logged_function(logger)
    def asset_upload_url(self, release_id: str) -> str:
        """
        Get the correct upload url for a release
        https://gitea.com/api/swagger#/repository/repoCreateReleaseAttachment
        :param release_id: ID of the release to upload to
        """
        return self.create_api_url(
            endpoint=f"/repos/{self.owner}/{self.repo_name}/releases/{release_id}/assets",
        )

    @logged_function(logger)
    def upload_release_asset(
        self,
        release_id: int,
        file: str,
        label: str | None = None,  # noqa: ARG002
    ) -> bool:
        """
        Upload an asset to an existing release
        https://gitea.com/api/swagger#/repository/repoCreateReleaseAttachment
        :param release_id: ID of the release to upload to
        :param file: Path of the file to upload
        :param label: this parameter has no effect

        :return: The status of the request
        """
        url = self.asset_upload_url(release_id)

        with open(file, "rb") as attachment:
            name = os.path.basename(file)
            content_type = "application/octet-stream"
            response = self.session.post(
                url,
                params={"name": name},
                data={},
                files={
                    "attachment": (
                        name,
                        attachment,
                        content_type,
                    ),
                },
            )

            # Raise an error if the request was not successful
            response.raise_for_status()

        logger.info(
            "Successfully uploaded %s to Gitea, url: %s, status code: %s",
            file,
            response.url,
            response.status_code,
        )

        return True

    @logged_function(logger)
    def upload_dists(self, tag: str, dist_glob: str) -> int:
        """
        Upload distributions to a release
        :param tag: Tag to upload for
        :param path: Path to the dist directory

        :return: The number of distributions successfully uploaded
        """
        # Find the release corresponding to this tag
        release_id = self.get_release_id_by_tag(tag=tag)
        if not release_id:
            logger.warning("No release corresponds to tag %s, can't upload dists", tag)
            return 0

        # Upload assets
        n_succeeded = 0
        for file_path in (
            f for f in glob.glob(dist_glob, recursive=True) if os.path.isfile(f)
        ):
            try:
                self.upload_release_asset(release_id, file_path)
                n_succeeded += 1
            except HTTPError:  # noqa: PERF203
                logger.exception("error uploading asset %s", file_path)

        return n_succeeded

    def remote_url(self, use_token: bool = True) -> str:
        """Get the remote url including the token for authentication if requested"""
        if not (self.token and use_token):
            return self._remote_url

        return self.create_server_url(
            auth=self.token,
            path=f"{self.owner}/{self.repo_name}.git",
        )

    def commit_hash_url(self, commit_hash: str) -> str:
        return self.create_repo_url(repo_path=f"/commit/{commit_hash}")

    def issue_url(self, issue_num: str | int) -> str:
        # Strips off any character prefix like '#' that usually exists
        if isinstance(issue_num, str) and (
            match := regexp(r"(\d+)$").search(issue_num)
        ):
            try:
                issue_num = int(match.group(1))
            except ValueError:
                return ""

        if isinstance(issue_num, int):
            return self.create_repo_url(repo_path=f"/issues/{issue_num}")

        return ""

    def pull_request_url(self, pr_number: str | int) -> str:
        # Strips off any character prefix like '#' that usually exists
        if isinstance(pr_number, str) and (
            match := regexp(r"(\d+)$").search(pr_number)
        ):
            try:
                pr_number = int(match.group(1))
            except ValueError:
                return ""

        if isinstance(pr_number, int):
            return self.create_repo_url(repo_path=f"/pulls/{pr_number}")

        return ""

    def create_release_url(self, tag: str = "") -> str:
        tag_str = tag.strip()
        tag_path = f"tag/{tag_str}" if tag_str else ""
        return self.create_repo_url(repo_path=f"releases/{tag_path}")

    @staticmethod
    def format_w_official_vcs_name(format_str: str) -> str:
        if "%s" in format_str:
            return format_str % Gitea.OFFICIAL_NAME

        if "{}" in format_str:
            return format_str.format(Gitea.OFFICIAL_NAME)

        if "{vcs_name}" in format_str:
            return format_str.format(vcs_name=Gitea.OFFICIAL_NAME)

        return format_str

    def get_changelog_context_filters(self) -> tuple[Callable[..., Any], ...]:
        return (
            self.create_server_url,
            self.create_repo_url,
            self.commit_hash_url,
            self.issue_url,
            self.pull_request_url,
            self.create_release_url,
            self.format_w_official_vcs_name,
        )


RemoteHvcsBase.register(Gitea)



from __future__ import annotations

# TODO: This tests for the main options that will help configuring a template,
# but not all of them. The testing can be expanded to cover all the options later.
# It's not super essential as Jinja2 does most of the testing, we're just checking
# that we can properly set the right strings in the template environment.
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from semantic_release.changelog.template import environment

if TYPE_CHECKING:
    from typing import Any

EXAMPLE_TEMPLATE_FORMAT_STR = """
<h1>This is an example template document</h1>

<h2>The title is {variable_start_string} title | upper {variable_end_string}</h2>
{comment_start_string}- This text should not appear {comment_end_string}
{block_start_string}- for subject in subjects {block_end_string}
<p>This is a paragraph about {variable_start_string} subject {variable_end_string}</p>
{block_start_string}- endfor {block_end_string}"""


@pytest.mark.parametrize(
    "format_map",
    [
        {
            "block_start_string": "{%",
            "block_end_string": "%}",
            "variable_start_string": "{{",
            "variable_end_string": "}}",
            "comment_start_string": "{#",
            "comment_end_string": "#}",
        },
        {
            "block_start_string": "{[",
            "block_end_string": "]}",
            "variable_start_string": "{{",
            "variable_end_string": "}}",
            "comment_start_string": "/*",
            "comment_end_string": "*/",
        },
    ],
)
@pytest.mark.parametrize(
    "subjects", [("dogs", "cats"), ("stocks", "finance", "politics")]
)
def test_template_env_configurable(format_map: dict[str, Any], subjects: tuple[str]):
    template_as_str = EXAMPLE_TEMPLATE_FORMAT_STR.format_map(format_map)
    env = environment(**format_map)
    template = env.from_string(template_as_str)

    title = "important"
    newline = "\n"
    expected_result = dedent(
        f"""
        <h1>This is an example template document</h1>

        <h2>The title is {title.upper()}</h2>
        {(newline + " " * 8).join(f'<p>This is a paragraph about {subject}</p>' for subject in subjects)}"""  # noqa: E501
    )
    actual_result = template.render(title="important", subjects=subjects)

    assert expected_result == actual_result
